""" REST views for PidXPath collection
"""
from rest_framework.permissions import IsAuthenticated
from rest_framework_mongoengine.generics import (
    ListCreateAPIView,
    RetrieveUpdateDestroyAPIView,
)

from core_linked_records_app.components.pid_xpath.models import PidXpath
from core_linked_records_app.rest.pid_xpath.serializers import PidXpathSerializer


class PidXpathListView(ListCreateAPIView):
    """View for listing and creating `PidXPath` objects"""

    permission_classes = (IsAuthenticated,)

    queryset = PidXpath.objects.all()
    serializer_class = PidXpathSerializer


class PidXpathDetailView(RetrieveUpdateDestroyAPIView):
    """View for listing and updating a single `PidXPath` object"""

    permission_classes = (IsAuthenticated,)

    queryset = PidXpath.objects.all()
    serializer_class = PidXpathSerializer



""" Unit tests for core_linked_records_app.utils.xml
"""
from unittest import TestCase


class TestGetXpathFromDotNotation(TestCase):
    def test_attributes_are_not_prepended_for_namespaces(self):
        pass

    def test_dot_notation_transformed_correctly(self):
        pass


class TestGetXpathWithTargetNamespace(TestCase):
    pass


class TestSetValueAtXpath(TestCase):
    def test_invalid_xpath_raises_error(self):
        pass

    def test_element_xpath_replaced(self):
        pass

    def test_attribute_xpath_replaced(self):
        pass


class TestGetValueAtXpath(TestCase):
    def test_not_existing_xpath_raises_error(self):
        pass

    def test_invalid_xpath_raises_error(self):
        pass

    def test_attribute_xpath_returns_value(self):
        pass

    def test_element_xpath_returns_value(self):
        pass



""" REST views for the query API
"""
from rest_framework import status
from rest_framework.response import Response

from core_linked_records_app import settings
from core_linked_records_app.components.pid_xpath import api as pid_xpath_api
from core_linked_records_app.utils.dict import get_dict_value_from_key_list
from core_main_app.components.data import api as data_api
from core_main_app.rest.data.abstract_views import AbstractExecuteLocalQueryView


class ExecuteLocalPIDQueryView(AbstractExecuteLocalQueryView):
    def build_query(
        self, query, workspaces=None, templates=None, options=None, title=None
    ):
        """Build the query by adding an extra filter to limit to document with
        PID fields.

        Args:
            query:
            workspaces:
            templates:
            options:
            title:

        Returns:
            Query with additional PID filter
        """
        pid_xpath_list = [
            pid_xpath_object.xpath
            for pid_xpath_object in pid_xpath_api.get_all(self.request)
        ] + [settings.PID_XPATH]
        pid_query = {
            "$and": [
                {
                    "$or": [
                        {f"dict_content.{pid_xpath}": {"$exists": 1}}
                        for pid_xpath in pid_xpath_list
                    ]
                }
            ]
        }
        query = super().build_query(query, workspaces, templates, options, title)

        if "$and" in query.keys():
            pid_query["$and"] += query["$and"]
        else:
            pid_query["$and"].append(query)

        return pid_query

    def execute_raw_query(self, raw_query, order_by_field):
        """Execute the raw query in database

        Args:

            raw_query: Query to execute
            order_by_field:

        Returns:
            Results of the query
        """
        pid_list = list()
        data_list = data_api.execute_query(raw_query, self.request.user)

        for data in data_list:
            pid_xpath_object = pid_xpath_api.get_by_template_id(
                data.template.pk, self.request
            )
            pid_xpath = pid_xpath_object.xpath

            data_pid = get_dict_value_from_key_list(
                data.dict_content,
                pid_xpath.split("."),
            )

            if not data_pid:
                continue

            pid_list.append({"pid": data_pid})

        return pid_list

    def build_response(self, data_list):
        """Build the list of PIDs

        Args:

            data_list: List of data

        Returns:

            Paginated list of data
        """
        # Send list of PID as JSON
        return Response(
            [data["pid"] for data in data_list],
            status=status.HTTP_200_OK,
        )


if (
    "core_oaipmh_harvester_app" in settings.INSTALLED_APPS
    and "core_explore_oaipmh_app" in settings.INSTALLED_APPS
):
    from core_oaipmh_harvester_app.components.oai_record import api as oai_record_api
    from core_explore_oaipmh_app.rest.query.views import ExecuteQueryView

    class ExecuteOaiPmhPIDQueryView(ExecuteQueryView):
        pid_xpath_list = None

        def build_query(self, query, templates, registries):
            """Build the query by adding an extra filter to limit to document with
            PID fields.

            Args:
                query:
                templates:
                registries:

            Returns:
                Query with additional PID filter
            """
            self.pid_xpath_list = [
                pid_xpath_object.xpath
                for pid_xpath_object in pid_xpath_api.get_all(self.request)
            ] + [settings.PID_XPATH]
            pid_query = {
                "$and": [
                    {
                        "$or": [
                            {f"dict_content.{pid_xpath}": {"$exists": 1}}
                            for pid_xpath in self.pid_xpath_list
                        ]
                    }
                ]
            }
            query = super().build_query(query, templates, registries)

            if "$and" in query.keys():
                pid_query["$and"] += query["$and"]
            else:
                pid_query["$and"].append(query)

            return pid_query

        def execute_raw_query(self, raw_query, order_by_field):
            """Execute the raw query in database

            Args:

                raw_query: Query to execute
                order_by_field:

            Returns:
                Results of the query
            """
            if self.pid_xpath_list is None:
                raise Exception("Undefined PID xpath from remote server.")

            pid_list = list()
            data_list = oai_record_api.execute_query(raw_query, self.request.user)

            for data in data_list:
                pid_xpath_object = pid_xpath_api.get_by_template_id(
                    data.harvester_metadata_format.template.pk, self.request
                )
                pid_xpath = pid_xpath_object.xpath

                data_pid = get_dict_value_from_key_list(
                    data["dict_content"],
                    pid_xpath.split("."),
                )

                if not data_pid:
                    continue

                pid_list.append({"pid": data_pid})

            return pid_list

        def build_response(self, data_list):
            """Build the list of PIDs

            Args:

                data_list: List of data

            Returns:

                Paginated list of data
            """
            # Send list of PID as JSON
            return Response(
                [data["pid"] for data in data_list],
                status=status.HTTP_200_OK,
            )



""" Unit tests for core_linked_records_app.components.pid_xpath.models
"""
from unittest import TestCase

from mongoengine import errors as mongoengine_errors
from unittest.mock import patch

from core_linked_records_app.components.pid_xpath.models import PidXpath
from core_main_app.commons.exceptions import ModelError


class TestPidXpathGetAll(TestCase):
    @patch.object(PidXpath, "objects")
    def test_pid_xpath_all_failure_raises_model_error(self, mock_pid_xpath):
        mock_pid_xpath.all.side_effect = Exception("mock_pid_xpath_all_exception")

        with self.assertRaises(ModelError):
            PidXpath.get_all()

    @patch.object(PidXpath, "objects")
    def test_returns_pid_xpath_all_output(self, mock_pid_xpath):
        expected_result = "mock_pid_xpath_all"
        mock_pid_xpath.all.return_value = expected_result

        self.assertEquals(PidXpath.get_all(), expected_result)


class TestPidXpathGetByTemplateList(TestCase):
    @patch.object(PidXpath, "objects")
    def test_pid_xpath_filter_failure_raises_model_error(self, mock_pid_xpath):
        mock_pid_xpath.filter.side_effect = Exception("mock_pid_xpath_filter_exception")

        with self.assertRaises(ModelError):
            PidXpath.get_all_by_template_list("mock_template_list")

    @patch.object(PidXpath, "objects")
    def test_returns_pid_xpath_filter_output(self, mock_pid_xpath):
        expected_result = "mock_pid_xpath_filter"
        mock_pid_xpath.filter.return_value = expected_result

        self.assertEquals(
            PidXpath.get_all_by_template_list("mock_template_list"), expected_result
        )


class TestPidXpathGetByTemplateId(TestCase):
    @patch.object(PidXpath, "objects")
    def test_pid_xpath_get_does_not_exists_returns_none(self, mock_pid_xpath):
        mock_pid_xpath.get.side_effect = mongoengine_errors.DoesNotExist()

        self.assertIsNone(PidXpath.get_by_template_id("mock_template_id"))

    @patch.object(PidXpath, "objects")
    def test_pid_xpath_get_raises_model_error(self, mock_pid_xpath):
        mock_pid_xpath.get.side_effect = Exception("mock_pid_xpath_get_exception")

        with self.assertRaises(ModelError):
            PidXpath.get_by_template_id("mock_template_id")

    @patch.object(PidXpath, "objects")
    def test_returns_pid_xpath_get_output(self, mock_pid_xpath):
        expected_result = "mock_pid_xpath_get"
        mock_pid_xpath.get.return_value = expected_result

        self.assertEquals(
            PidXpath.get_by_template_id("mock_template_id"), expected_result
        )



""" Local resolver API
"""
from logging import getLogger

from core_linked_records_app import settings
from core_linked_records_app.components.pid_xpath import api as pid_xpath_api
from core_linked_records_app.utils.dict import get_dict_value_from_key_list
from core_main_app.commons.exceptions import ApiError, DoesNotExist
from core_main_app.components.data import api as data_api

logger = getLogger(__name__)


def get_data_by_pid(pid, request):
    """Return data object with the given pid.

    Parameters:
        pid:
        request: HttpRequest

    Returns: data object
    """
    try:
        pid_xpath_list = [
            pid_xpath_object.xpath
            for pid_xpath_object in pid_xpath_api.get_all(request)
        ] + [settings.PID_XPATH]
        json_pid_xpaths = [f"dict_content.{pid_xpath}" for pid_xpath in pid_xpath_list]
        query_result = data_api.execute_query(
            {"$or": [{json_pid_xpath: pid} for json_pid_xpath in json_pid_xpaths]},
            request.user,
        )
        query_result_length = len(query_result)
    except Exception as exc:
        error_message = (
            f"An error occurred while looking up data assigned to PID '{pid}'"
        )

        logger.error(f"{error_message}: {str(exc)}")
        raise ApiError(f"{error_message}.")

    if query_result_length == 0:
        raise DoesNotExist("PID is not attached to any data.")
    elif query_result_length != 1:
        raise ApiError("PID must be unique.")
    else:
        return query_result[0]


def get_pids_for_data_list(data_id_list, request):
    """Retrieve PID matching the document list provided.

    Args:
        data_id_list:
        request: HttpRequest

    Returns:
    """
    try:
        # Retrieve the document passed as input and extra the PID field.
        data_list = data_api.get_by_id_list(data_id_list, request.user)

        # Build the list of PID from the document and the pid_xpath retrieve from the
        # `PidSettings`.
        pid_xpath_list = list()

        for data in data_list:
            pid_xpath_object = pid_xpath_api.get_by_template_id(
                data.template.pk, request
            )
            pid_xpath_list.append((data, pid_xpath_object.xpath))

        pid_list = [
            get_dict_value_from_key_list(
                data.dict_content,
                pid_xpath.split("."),
            )
            for data, pid_xpath in pid_xpath_list
        ]

        # Returns a list of PID available from the list.
        return [pid for pid in pid_list if pid is not None]
    except Exception as exc:
        error_message = f"An error occurred while retrieving PIDs for data list"

        logger.error(f"{error_message}: {str(exc)}")
        raise ApiError(f"{error_message}.")


def get_pid_for_data(data_id, request):
    """Retrieve PID matching the document ID provided.

    Args:
        data_id:
        request: HttpRequest

    Returns:
    """
    try:
        # Retrieve the document passed as input and extra the PID field.
        data = data_api.get_by_id(data_id, request.user)

        # Return PID value from the document and the pid_xpath retrieved from
        # `PidSettings`
        pid_xpath_object = pid_xpath_api.get_by_template_id(data.template.pk, request)
        pid_xpath = pid_xpath_object.xpath

        return get_dict_value_from_key_list(data.dict_content, pid_xpath.split("."))
    except Exception as exc:
        error_message = (
            f"An error occurred while looking up PID assigned to data '{data_id}'"
        )

        logger.error(f"{error_message}: {str(exc)}")
        raise ApiError(f"{error_message}.")



""" Unit tests for core_linked_records_app.components.oai_record.api
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.oai_record import api as oai_record_api
from core_linked_records_app.components.pid_xpath import api as pid_xpath_api
from core_main_app.commons.exceptions import ApiError
from core_oaipmh_harvester_app.components.oai_record import api as oai_record_data
from tests import mocks


class TestGetPidForData(TestCase):
    @patch.object(oai_record_data, "get_by_id")
    def test_get_by_id_failure_raises_api_error(self, mock_get_by_id):
        mock_get_by_id.side_effect = Exception("mock_get_by_id_exception")

        with self.assertRaises(ApiError):
            oai_record_api.get_pid_for_data("mock_id", mocks.MockRequest())

    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(oai_record_data, "get_by_id")
    def test_get_by_template_id_failure_raises_api_error(
        self, mock_get_by_id, mock_get_by_template_id
    ):
        mock_get_by_id.return_value = mocks.MockData()
        mock_get_by_template_id.side_effect = Exception(
            "mock_get_by_template_id_exception"
        )

        with self.assertRaises(ApiError):
            oai_record_api.get_pid_for_data("mock_id", mocks.MockRequest())

    @patch.object(oai_record_api, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(oai_record_data, "get_by_id")
    def test_get_dict_value_from_key_list_failure_raises_api_error(
        self, mock_get_by_id, mock_get_by_template_id, mock_get_dict_value_from_key_list
    ):
        mock_get_by_id.return_value = mocks.MockData()
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.side_effect = Exception(
            "mock_get_dict_value_from_key_list_exception"
        )

        with self.assertRaises(ApiError):
            oai_record_api.get_pid_for_data("mock_id", mocks.MockRequest())

    @patch.object(oai_record_api, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(oai_record_data, "get_by_id")
    def test_returns_get_dict_value_from_key_list(
        self, mock_get_by_id, mock_get_by_template_id, mock_get_dict_value_from_key_list
    ):
        expected_result = "mock_get_pid_for_data"
        mock_get_by_id.return_value = mocks.MockData()
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.return_value = expected_result

        self.assertEquals(
            oai_record_api.get_pid_for_data("mock_id", mocks.MockRequest()),
            expected_result,
        )



""" Unit tests for core_linked_records_app.components.blob.api
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.components.local_id import api as local_id_api
from core_linked_records_app.components.local_id.models import LocalId
from core_main_app.commons import exceptions
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests import mocks


class TestGetBlobByPid(TestCase):
    def setUp(self):
        self.user = create_mock_user("1")

    @patch.object(local_id_api, "get_by_name")
    def test_get_by_name_error_raises_api_error(self, mock_get_by_name):
        mock_get_by_name.side_effect = Exception("mock_get_by_name_exception")
        mock_valid_pid = "https://websi.te/provider/record"

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(local_id_api, "get_by_name")
    def test_undefined_classpath_raises_does_not_exist(self, mock_get_by_name):
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(record_name=mock_valid_pid)

        with self.assertRaises(exceptions.DoesNotExist):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(blob_api, "import_module")
    @patch.object(local_id_api, "get_by_name")
    def test_failed_import_raises_api_error(self, mock_get_by_name, mock_import_module):
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.side_effect = Exception("mock_import_module_exception")

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_api, "get_by_name")
    def test_get_by_id_exception_raises_api_error(
        self, mock_get_by_name, mock_import_module, mock_getattr
    ):
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.return_value = None
        mock_getattr.side_effect = Exception("mock_getattr_exception")

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_blob_by_pid(mock_valid_pid, self.user)

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_api, "get_by_name")
    def test_returns_expected_get_by_id_output(
        self, mock_get_by_name, mock_import_module, mock_getattr
    ):
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.return_value = None
        mock_getattr.return_value = mocks.MockModule()

        result = blob_api.get_blob_by_pid(mock_valid_pid, self.user)
        self.assertEquals(result, mocks.MockModule().get_by_id())


class TestGetPidForBlob(TestCase):
    @patch.object(blob_api, "get_api_path_from_object")
    def test_get_api_path_from_object_exception_raises_api_error(
        self, mock_get_api_path_from_object
    ):
        mock_get_api_path_from_object.side_effect = Exception(
            "mock_get_api_path_from_object_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_pid_for_blob("mock_blob_id")

    @patch.object(local_id_api, "get_by_class_and_id")
    @patch.object(blob_api, "get_api_path_from_object")
    def test_get_by_class_and_id_exception_raises_api_error(
        self, mock_get_api_path_from_object, mock_get_by_class_and_id
    ):
        mock_get_api_path_from_object.return_value = ""
        mock_get_by_class_and_id.side_effect = Exception(
            "mock_get_by_class_and_id_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            blob_api.get_pid_for_blob("mock_blob_id")

    @patch.object(local_id_api, "get_by_class_and_id")
    @patch.object(blob_api, "get_api_path_from_object")
    def test_returns_get_by_class_and_id_output(
        self, mock_get_api_path_from_object, mock_get_by_class_and_id
    ):
        mock_pid_value = "mock_pid_value"
        mock_get_api_path_from_object.return_value = ""
        mock_get_by_class_and_id.return_value = mock_pid_value

        result = blob_api.get_pid_for_blob("mock_blob_id")
        self.assertEquals(result, mock_pid_value)


class TestSetPidForBlob(TestCase):
    @patch.object(blob_api, "get_pid_for_blob")
    def test_get_pid_for_blob_exception_raises_api_error(self, mock_get_pid_for_blob):
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = Exception("mock_get_pid_for_blob_exception")

        with self.assertRaises(exceptions.ApiError):
            blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)

    @patch.object(local_id_api, "get_by_name")
    @patch.object(blob_api, "get_pid_for_blob")
    def test_get_by_name_exception_raises_api_error(
        self, mock_get_pid_for_blob, mock_get_by_name
    ):
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_by_name.side_effect = Exception("mock_get_by_name_exception")

        with self.assertRaises(exceptions.ApiError):
            blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)

    @patch.object(blob_api, "get_api_path_from_object")
    @patch.object(local_id_api, "get_by_name")
    @patch.object(blob_api, "get_pid_for_blob")
    def test_new_local_id_get_api_path_from_object_exception_raises_api_error(
        self, mock_get_pid_for_blob, mock_get_by_name, mock_get_api_path_from_object
    ):
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_by_name.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_api_path_from_object.side_effect = Exception(
            "mock_get_api_path_from_object_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)

    @patch.object(blob_api, "get_api_path_from_object")
    @patch.object(local_id_api, "get_by_name")
    @patch.object(blob_api, "get_pid_for_blob")
    def test_edit_local_id_get_api_path_from_object_exception_raises_api_error(
        self, mock_get_pid_for_blob, mock_get_by_name, mock_get_api_path_from_object
    ):
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_by_name.return_value = mocks.MockLocalId()
        mock_get_api_path_from_object.side_effect = Exception(
            "mock_get_api_path_from_object_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)

    @patch.object(local_id_api, "insert")
    @patch.object(blob_api, "get_api_path_from_object")
    @patch.object(local_id_api, "get_by_name")
    @patch.object(blob_api, "get_pid_for_blob")
    def test_new_local_id_insert_exception_raises_api_error(
        self,
        mock_get_pid_for_blob,
        mock_get_by_name,
        mock_get_api_path_from_object,
        mock_insert,
    ):
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_by_name.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_api_path_from_object.return_value = "mock_api_path"
        mock_insert.side_effect = Exception("mock_side_effect_exception")

        with self.assertRaises(exceptions.ApiError):
            blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)

    @patch.object(local_id_api, "insert")
    @patch.object(blob_api, "get_api_path_from_object")
    @patch.object(local_id_api, "get_by_name")
    @patch.object(blob_api, "get_pid_for_blob")
    def test_edit_local_id_insert_exception_raises_api_error(
        self,
        mock_get_pid_for_blob,
        mock_get_by_name,
        mock_get_api_path_from_object,
        mock_insert,
    ):
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_by_name.return_value = mocks.MockLocalId()
        mock_get_api_path_from_object.return_value = "mock_api_path"
        mock_insert.side_effect = Exception("mock_side_effect_exception")

        with self.assertRaises(exceptions.ApiError):
            blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)

    @patch.object(local_id_api, "insert")
    @patch.object(blob_api, "get_api_path_from_object")
    @patch.object(local_id_api, "get_by_name")
    @patch.object(blob_api, "get_pid_for_blob")
    def test_new_local_id_returns_insert_output(
        self,
        mock_get_pid_for_blob,
        mock_get_by_name,
        mock_get_api_path_from_object,
        mock_insert,
    ):
        mock_insert_result = "mock_local_id_object"
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_by_name.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_api_path_from_object.return_value = "mock_api_path"
        mock_insert.return_value = mock_insert_result

        result = blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)
        self.assertEquals(result, mock_insert_result)

    @patch.object(local_id_api, "insert")
    @patch.object(blob_api, "get_api_path_from_object")
    @patch.object(local_id_api, "get_by_name")
    @patch.object(blob_api, "get_pid_for_blob")
    def test_edit_local_id_returns_insert_output(
        self,
        mock_get_pid_for_blob,
        mock_get_by_name,
        mock_get_api_path_from_object,
        mock_insert,
    ):

        mock_insert_result = "mock_local_id_object"
        mock_blob_id = "mock_blob_id"
        mock_blob_pid = "mock_blob_pid"
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist("does_not_exist")
        mock_get_by_name.return_value = mocks.MockLocalId()
        mock_get_api_path_from_object.return_value = "mock_api_path"
        mock_insert.return_value = mock_insert_result

        result = blob_api.set_pid_for_blob(mock_blob_id, mock_blob_pid)
        self.assertEquals(result, mock_insert_result)



""" Linked Records Settings
"""
from django_mongoengine import Document
from mongoengine import errors as mongoengine_errors
from mongoengine import fields

from core_linked_records_app import settings
from core_main_app.commons import exceptions


class PidSettings(Document):
    auto_set_pid = fields.BooleanField(default=settings.AUTO_SET_PID)

    @staticmethod
    def get():
        try:
            return PidSettings.objects.first()
        except mongoengine_errors.DoesNotExist:
            return None
        except Exception as exc:
            raise exceptions.ModelError(str(exc))



""" Unit tests for core_linked_records_app.system.api
"""
from unittest import TestCase


class TestIsPidDefinedForDocument(TestCase):
    def test_wrong_document_id_raise_error(self):
        pass

    def test_undefined_pid_returns_false(self):
        pass

    def test_duplicate_pid_returns_false(self):
        pass

    def test_defined_pid_for_other_document_returns_false(self):
        pass

    def test_defined_pid_for_current_document_returns_true(self):
        pass


class TestIsPidDefined(TestCase):
    def test_get_data_by_pid_fails_with_unexpected_error_raises_error(self):
        pass

    def test_get_data_by_pid_fails_with_expected_error_returns_false(self):
        pass

    def test_get_data_by_pid_succeeds_return_true(self):
        pass


class TestGetDataByPid(TestCase):
    def test_query_returns_no_results_raises_error(self):
        pass

    def test_query_returns_several_results_raises_error(self):
        pass

    def test_query_returns_single_result_returns_result(self):
        pass


class TestGetPidForData(TestCase):
    def test_not_existing_pid_returns_none(self):
        pass

    def test_existing_pid_returns_pid(self):
        pass



""" Unit tests for core_linked_records_app.utils.dict
"""
from unittest import TestCase


class TestGetDictValueFromKeyList(TestCase):
    def test_empty_key_list_returns_dictionary(self):
        pass

    def test_non_existing_key_returns_none(self):
        pass

    def test_existing_key_returns_dict_content(self):
        pass



""" Local handle model
"""
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors

from core_main_app.commons import exceptions
from core_main_app.commons.regex import NOT_EMPTY_OR_WHITESPACES


class LocalId(Document):
    """Handle object"""

    record_name = fields.StringField(
        blank=False, unique=True, regex=NOT_EMPTY_OR_WHITESPACES
    )
    record_object_class = fields.StringField(blank=True)
    record_object_id = fields.StringField(blank=True)

    @staticmethod
    def get_by_name(record_name):
        try:
            return LocalId.objects.get(record_name=record_name)
        except mongoengine_errors.DoesNotExist as dne:
            raise exceptions.DoesNotExist(str(dne))
        except Exception as exc:
            raise exceptions.ModelError(str(exc))

    @staticmethod
    def get_by_class_and_id(record_object_class, record_object_id):
        try:
            return LocalId.objects.get(
                record_object_class=record_object_class,
                record_object_id=record_object_id,
            )
        except mongoengine_errors.DoesNotExist as dne:
            raise exceptions.DoesNotExist(str(dne))
        except Exception as exc:
            raise exceptions.ModelError(str(exc))

    @staticmethod
    def upsert(local_id_object):
        try:
            return local_id_object.save()
        except mongoengine_errors.NotUniqueError as nue:
            raise exceptions.NotUniqueError(str(nue))
        except Exception as exc:
            raise exceptions.ModelError(str(exc))



""" Serializer classes for LocalID object
"""
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_linked_records_app.components.local_id.models import LocalId


class LocalIdSerializer(DocumentSerializer):
    """LocalId serializer"""

    class Meta(object):
        model = LocalId
        fields = "__all__"
        read_only_fields = ["record_name", "record_object_class", "record_object_id"]



""" Unit tests for core_linked_records_app.utils.providers.handle_net
"""
from unittest import TestCase


class TestHandleNetSystemUpdateResponseContent(TestCase):
    def test_returns_dict_with_message_key(self):
        pass

    def test_returns_dict_with_url_key(self):
        pass


class TestHandleNetSystemEncodeToken(TestCase):
    def test_returned_base64_encoding(self):
        pass


class TestHandleNetSystemGet(TestCase):
    def test_record_url_called(self):
        pass

    def test_returns_dict_with_message_key(self):
        pass

    def test_returns_dict_with_url_key(self):
        pass


class TestHandleNetSystemCreate(TestCase):
    def test_record_not_none_calls_existing_record_url(self):
        pass

    def test_record_not_none_returns_dict_with_message_key(self):
        pass

    def test_record_not_none_returns_dict_with_url_key(self):
        pass

    def test_record_none_calls_new_record_url(self):
        pass

    def test_record_none_calls_update_method(self):
        pass


class TestHandleNetSystemUpdate(TestCase):
    def test_send_put_request_on_record_url(self):
        pass

    def test_returns_dict_with_message_key(self):
        pass

    def test_returns_dict_with_url_key(self):
        pass


class TestHandleNetSystemDelete(TestCase):
    def test_send_delete_request_on_record_url(self):
        pass

    def test_returns_dict_with_message_key(self):
        pass

    def test_returns_dict_with_url_key(self):
        pass



""" Unit tests for core_linked_records_app.components.local_id.api
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.local_id import api as local_id_api
from core_linked_records_app.components.local_id.models import LocalId
from core_main_app.commons import exceptions


class TestGetByName(TestCase):
    @patch.object(LocalId, "get_by_name")
    def test_get_by_name_failure_raises_api_error(self, mock_get_by_name):
        mock_get_by_name.side_effect = Exception("mock_get_by_name_exception")

        with self.assertRaises(exceptions.ApiError):
            local_id_api.get_by_name("mock_name")

    @patch.object(LocalId, "get_by_name")
    def test_returns_get_by_name_output(self, mock_get_by_name):
        expected_result = "mock_get_by_name"
        mock_get_by_name.return_value = expected_result

        self.assertEquals(local_id_api.get_by_name("mock_name"), expected_result)


class TestGetByClassAndId(TestCase):
    @patch.object(LocalId, "get_by_class_and_id")
    def test_get_by_class_and_id_failure_raises_api_error(
        self, mock_get_by_class_and_id
    ):
        mock_get_by_class_and_id.side_effect = Exception(
            "mock_get_by_class_and_id_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            local_id_api.get_by_class_and_id("class", "id")

    @patch.object(LocalId, "get_by_class_and_id")
    def test_returns_get_by_class_and_id_output(self, mock_get_by_class_and_id):
        expected_result = "mock_get_by_class_and_id"
        mock_get_by_class_and_id.return_value = expected_result

        self.assertEquals(
            local_id_api.get_by_class_and_id("class", "id"), expected_result
        )


class TestInsert(TestCase):
    def setUp(self) -> None:
        self.mock_local_id = LocalId(record_name="mock_record_name")

    @patch.object(LocalId, "upsert")
    def test_save_failure_raises_api_error(self, mock_save):
        mock_save.side_effect = Exception("mock_save_general_exception")

        with self.assertRaises(exceptions.ApiError):
            local_id_api.insert(self.mock_local_id)

    @patch.object(LocalId, "upsert")
    def test_returns_save_output(self, mock_save):
        expected_result = "mock_save"
        mock_save.return_value = expected_result

        self.assertEquals(local_id_api.insert(self.mock_local_id), expected_result)


class TestDelete(TestCase):
    def setUp(self) -> None:
        self.mock_local_id = LocalId(record_name="mock_record_name")

    @patch.object(LocalId, "delete")
    def test_delete_raises_api_error(self, mock_delete):
        mock_delete.side_effect = Exception("mock_delete_exception")

        with self.assertRaises(exceptions.ApiError):
            local_id_api.delete(self.mock_local_id)

    @patch.object(LocalId, "delete")
    def test_returns_delete_output(self, mock_delete):
        expected_result = "mock_delete"
        mock_delete.return_value = expected_result

        self.assertEquals(local_id_api.delete(self.mock_local_id), expected_result)



""" System API for core_linked_records
"""
import logging
from rest_framework import status

from core_linked_records_app import settings
from core_linked_records_app.components.pid_xpath.models import PidXpath
from core_linked_records_app.utils.dict import get_dict_value_from_key_list
from core_linked_records_app.utils.providers import ProviderManager
from core_main_app.commons.exceptions import DoesNotExist, ApiError
from core_main_app.components.data.models import Data
from core_main_app.utils.requests_utils.requests_utils import send_delete_request

logger = logging.getLogger(__name__)


def is_pid_defined_for_document(pid, document_id):
    """Determine if a given PID match the document ID provided.

    Params:
        pid:
        document_id:

    Returns:
    """
    data = Data.get_by_id(document_id)

    pid_xpath_object = get_pid_xpath_by_template_id(data.template.pk)
    pid_xpath = pid_xpath_object.xpath

    json_pid_path = f"dict_content.{pid_xpath}"
    query_result = Data.execute_query({json_pid_path: pid}, order_by_field=[])

    return len(query_result) == 1 and query_result[0].pk == document_id


def is_pid_defined(pid):
    """Determine if a given PID already exists.

    Params:
        pid:

    Returns:
    """
    try:
        get_data_by_pid(pid)
        return True
    except (DoesNotExist, ApiError):
        return False


def get_data_by_pid(pid):
    """Return data object with the given pid.

    Parameters:
        pid:

    Returns: data object
    """
    pid_xpath_list = [
        pid_xpath_object.xpath for pid_xpath_object in PidXpath.get_all()
    ] + [settings.PID_XPATH]
    json_pid_xpaths = [f"dict_content.{pid_xpath}" for pid_xpath in pid_xpath_list]
    query_result = Data.execute_query(
        {"$or": [{json_pid_xpath: pid} for json_pid_xpath in json_pid_xpaths]},
        order_by_field=[],
    )
    query_result_length = len(query_result)

    if query_result_length == 0:
        raise DoesNotExist("PID is not attached to any data.")
    elif query_result_length != 1:
        raise ApiError("PID must be unique.")
    else:
        return query_result[0]


def get_pid_for_data(data_id):
    """Retrieve PID matching the document ID provided.

    Args:
        data_id:

    Returns:
    """
    # Retrieve the document passed as input and extra the PID field.
    data = Data.get_by_id(data_id)

    # Return PID value from the document and the PID_XPATH
    pid_xpath_object = get_pid_xpath_by_template_id(data.template.pk)
    pid_xpath = pid_xpath_object.xpath

    return get_dict_value_from_key_list(
        data["dict_content"],
        pid_xpath.split("."),
    )


def get_pid_xpath_by_template_id(template_id):
    """Retrieve XPath associated with a specific template ID

    Args:
        template_id: ObjectId

    Returns:
        PidXpath - PidXpath object, linking template ID and XPath
    """
    pid_xpath_object = PidXpath.get_by_template_id(template_id)

    if pid_xpath_object is None:
        return PidXpath(template=template_id, xpath=settings.PID_XPATH)
    else:
        return pid_xpath_object


def delete_pid_for_data(data):
    """Deletes the PID assigned to the data passed in parameter. If no PID has
    been assigned, the function simply exits.

    Args:
        data:
    """
    provider_manager = ProviderManager()
    previous_pid = get_pid_for_data(data.pk)

    if previous_pid is None:  # If there is no previous PID assigned
        logger.info(f"No PID assigned to the data {str(data.pk)}")
        return

    previous_provider_name = provider_manager.find_provider_from_pid(previous_pid)
    previous_provider = provider_manager.get(previous_provider_name)
    previous_pid_url = previous_pid.replace(
        previous_provider.provider_lookup_url, previous_provider.local_url
    )

    previous_pid_delete_response = send_delete_request(
        "%s?format=json" % previous_pid_url
    )

    # Log any error that happen during PID deletion
    if previous_pid_delete_response.status_code != status.HTTP_200_OK:
        logger.warning(
            "Deletion of PID %s returned %s"
            % (previous_pid, previous_pid_delete_response.status_code)
        )



""" Permission tests for core_linked_records_app.rest.pid.views
"""
from unittest import TestCase

from rest_framework import status
from unittest.mock import patch

from core_explore_common_app.components.query import api as query_api
from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.components.data import api as data_api
from core_linked_records_app.rest.pid import views as pid_views
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock
from tests import mocks


class TestRetrieveDataPidGet(TestCase):
    @patch.object(data_api, "get_pid_for_data")
    def test_anonymous_returns_200(self, mock_get_pid_for_data):
        mock_get_pid_for_data.return_value = "mock_pid"

        response = RequestMock.do_request_get(
            pid_views.RetrieveDataPIDView.as_view(),
            None,
            data={"data_id": "mock_data_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(data_api, "get_pid_for_data")
    def test_authenticated_returns_200(self, mock_get_pid_for_data):
        mock_user = create_mock_user("1")

        mock_get_pid_for_data.return_value = "mock_pid"

        response = RequestMock.do_request_get(
            pid_views.RetrieveDataPIDView.as_view(),
            mock_user,
            data={"data_id": "mock_data_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(data_api, "get_pid_for_data")
    def test_staff_returns_200(self, mock_get_pid_for_data):
        mock_user = create_mock_user("1", is_staff=True)

        mock_get_pid_for_data.return_value = "mock_pid"

        response = RequestMock.do_request_get(
            pid_views.RetrieveDataPIDView.as_view(),
            mock_user,
            data={"data_id": "mock_data_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestRetrieveBlobPidGet(TestCase):
    @patch.object(blob_api, "get_pid_for_blob")
    def test_anonymous_returns_(self, mock_get_pid_for_blob):
        mock_get_pid_for_blob.return_value = mocks.MockLocalId()

        response = RequestMock.do_request_get(
            pid_views.RetrieveBlobPIDView.as_view(),
            None,
            data={"blob_id": "mock_blob_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(blob_api, "get_pid_for_blob")
    def test_authenticated_returns_(self, mock_get_pid_for_blob):
        mock_user = create_mock_user("1")

        mock_get_pid_for_blob.return_value = mocks.MockLocalId()

        response = RequestMock.do_request_get(
            pid_views.RetrieveBlobPIDView.as_view(),
            mock_user,
            data={"blob_id": "mock_blob_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(blob_api, "get_pid_for_blob")
    def test_staff_returns_(self, mock_get_pid_for_blob):
        mock_user = create_mock_user("1", is_staff=True)

        mock_get_pid_for_blob.return_value = mocks.MockLocalId()

        response = RequestMock.do_request_get(
            pid_views.RetrieveBlobPIDView.as_view(),
            mock_user,
            data={"blob_id": "mock_blob_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestRetrieveListPidPost(TestCase):
    def setUp(self) -> None:
        self.mock_data_source = mocks.MockDataSource(
            authentication=mocks.MockAuthentication(
                type="oauth2", params={"access_token": "mock_access_token"}
            ),
            capabilities={"url_pid": True},
        )

    def send_post_request(self, mock_query_get_by_id, mock_send_get_request, user):
        mock_query_get_by_id.return_value = mocks.MockQuery(
            data_sources=[self.mock_data_source]
        )
        mock_send_get_request.return_value = mocks.MockResponse(json_data=[None])

        return RequestMock.do_request_post(
            pid_views.RetrieveListPIDView.as_view(),
            user,
            data={"query_id": "mock_query_id", "data_source_index": "0"},
            content_type=None,
        )

    @patch.object(pid_views, "oauth2_post_request")
    @patch.object(query_api, "get_by_id")
    def test_anonymous_returns_200(self, mock_query_get_by_id, mock_send_get_request):
        response = self.send_post_request(
            mock_query_get_by_id, mock_send_get_request, None
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(pid_views, "oauth2_post_request")
    @patch.object(query_api, "get_by_id")
    def test_authenticated_returns_200(
        self, mock_query_get_by_id, mock_send_get_request
    ):
        mock_user = create_mock_user("1")

        response = self.send_post_request(
            mock_query_get_by_id, mock_send_get_request, mock_user
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(pid_views, "oauth2_post_request")
    @patch.object(query_api, "get_by_id")
    def test_staff_returns_200(self, mock_query_get_by_id, mock_send_get_request):
        mock_user = create_mock_user("1", is_staff=True)

        response = self.send_post_request(
            mock_query_get_by_id, mock_send_get_request, mock_user
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)



""" Ajax views accessible by users.
"""
import json
from django.http import JsonResponse
from django.urls import reverse
from django.utils import timezone
from rest_framework import status
from rest_framework.views import APIView
from urllib.parse import urljoin

from core_explore_common_app.components.query import api as query_api
from core_explore_common_app.utils.protocols.oauth2 import (
    send_post_request as oauth2_post_request,
    send_get_request as oauth2_get_request,
)
from core_linked_records_app import settings
from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.components.data import api as data_api
from core_main_app.utils.requests_utils.requests_utils import (
    send_get_request,
)


class RetrieveDataPIDView(APIView):
    """Retrieve PIDs for a given data IDs."""

    def get(self, request):
        try:
            if "data_id" in request.GET:
                return JsonResponse(
                    {"pid": data_api.get_pid_for_data(request.GET["data_id"], request)}
                )
            elif (
                "core_oaipmh_harvester_app" in settings.INSTALLED_APPS
                and "core_explore_oaipmh_app" in settings.INSTALLED_APPS
                and "oai_data_id" in request.GET
            ):
                from core_linked_records_app.components.oai_record import (
                    api as oai_record_api,
                )

                return JsonResponse(
                    {
                        "pid": oai_record_api.get_pid_for_data(
                            request.GET["oai_data_id"], request
                        )
                    }
                )
            elif (
                "core_federated_search_app" in settings.INSTALLED_APPS
                and "fede_data_id" in request.GET
                and "fede_origin" in request.GET
            ):
                from core_federated_search_app.components.instance import (
                    api as instance_api,
                )

                fede_origin_keys = request.GET["fede_origin"].split("&")
                instance_name = fede_origin_keys[1].split("=")[1]
                instance = instance_api.get_by_name(instance_name)

                url_get_data = "%s?data_id=%s" % (
                    reverse(
                        "core_linked_records_retrieve_data_pid",
                    ),
                    request.GET["fede_data_id"],
                )
                data_response = oauth2_get_request(
                    urljoin(instance.endpoint, url_get_data), instance.access_token
                )
                return JsonResponse(json.loads(data_response.text))
            else:
                return JsonResponse(
                    {
                        "message": "Impossible to retrieve PID for data with the given "
                        "parameters"
                    },
                    status=status.HTTP_400_BAD_REQUEST,
                )
        except Exception as exc:
            return JsonResponse(
                {
                    "message": f"An unexpected exception occurred while retrieving data "
                    f"PID: {str(exc)}"
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )


class RetrieveBlobPIDView(APIView):
    """Retrieve PIDs for a given blob ID."""

    def get(self, request):
        if "blob_id" in request.GET:
            try:
                blob_pid = blob_api.get_pid_for_blob(request.GET["blob_id"])
                sub_url = reverse(
                    "core_linked_records_provider_record",
                    kwargs={
                        "provider": settings.ID_PROVIDER_SYSTEM_NAME,
                        "record": "",
                    },
                )

                return JsonResponse(
                    {"pid": f"{settings.SERVER_URI}{sub_url}{blob_pid.record_name}"}
                )
            except Exception as exc:
                return JsonResponse(
                    {
                        "message": f"An unexpected exception occurred while retrieving "
                        f"blob PID: {str(exc)}"
                    },
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                )
        else:
            return JsonResponse(
                {"message": "Missing parameter 'blob_id'."},
                status=status.HTTP_400_BAD_REQUEST,
            )


class RetrieveListPIDView(APIView):
    """Retrieve PIDs for a given list of data IDs."""

    def post(self, request):
        try:
            # FIXME duplicated code with core_explore_common.utils.query.send
            query = query_api.get_by_id(
                request.POST.get("query_id", None),
                request.user,
            )
            data_source = query.data_sources[
                int(request.POST.get("data_source_index", 0))
            ]

            # Build serialized query to send to data source
            json_query = {
                "query": query.content,
                "templates": json.dumps(
                    [
                        {"id": str(template.id), "hash": template.hash}
                        for template in query.templates
                    ]
                ),
                "options": json.dumps(data_source.query_options),
                "order_by_field": data_source.order_by_field,
            }

            if (
                getattr(data_source, "capabilities", False) is not False
                and "url_pid" not in data_source.capabilities.keys()
            ):
                return JsonResponse(
                    {"error": "The remote does not have PID capabilities."},
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                )

            if data_source.authentication.type == "session":
                response = send_get_request(
                    data_source.capabilities["url_pid"],
                    json=json_query,
                    cookies={
                        "sessionid": request.session.session_key,
                    },
                )
            elif data_source.authentication.type == "oauth2":
                response = oauth2_post_request(
                    data_source.capabilities["url_pid"],
                    json_query,
                    data_source.authentication.params["access_token"],
                    session_time_zone=timezone.get_current_timezone(),
                )
            else:
                raise Exception("Unknown authentication type.")

            if response.status_code == 200:
                return JsonResponse(
                    {"pids": [pid for pid in response.json() if pid is not None]},
                    status=response.status_code,
                )
            else:
                return JsonResponse(
                    {
                        "error": "Remote service answered with status code %d."
                        % response.status_code
                    },
                    status=response.status_code,
                )
        except Exception as exception:
            return JsonResponse(
                {"error": str(exception)}, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )



""" Unit tests for core_linked_records_app.components.pid_settings.api
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_linked_records_app.components.pid_settings.models import PidSettings
from core_main_app.commons.exceptions import ApiError


class TestUpsert(TestCase):
    def setUp(self) -> None:
        self.mock_pid_settings = PidSettings()

    @patch.object(PidSettings, "save")
    def test_save_failure_raises_api_error(self, mock_save):
        mock_save.side_effect = Exception("mock_save_exception")

        with self.assertRaises(ApiError):
            pid_settings_api.upsert(self.mock_pid_settings)

    @patch.object(PidSettings, "save")
    def test_returns_save_method_output(self, mock_save):
        expected_result = "mock_save"
        mock_save.return_value = expected_result

        self.assertEquals(
            pid_settings_api.upsert(self.mock_pid_settings), expected_result
        )


class TestGet(TestCase):
    @patch.object(PidSettings, "get")
    def test_save_failure_raises_api_error(self, mock_get):
        mock_get.side_effect = Exception("mock_get_exception")

        with self.assertRaises(ApiError):
            pid_settings_api.get()

    @patch.object(PidSettings, "get")
    def test_returns_save_method_output(self, mock_get):
        expected_result = "mock_get"
        mock_get.return_value = expected_result

        self.assertEquals(pid_settings_api.get(), expected_result)



""" Signals to trigger after Blob save
"""
from logging import getLogger

import json
from django.urls import reverse

from core_linked_records_app import settings
from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_main_app.commons.exceptions import CoreError, DoesNotExist
from core_main_app.components.blob.models import Blob
from core_main_app.utils.requests_utils.requests_utils import send_post_request
from signals_utils.signals.mongo import connector, signals

logger = getLogger(__name__)


def init():
    """Connect to Blob object events."""
    connector.connect(set_blob_pid, signals.post_save, sender=Blob)


def set_blob_pid(sender, document, **kwargs):
    try:
        if not pid_settings_api.get().auto_set_pid:
            return

        try:
            blob_api.get_pid_for_blob(str(document.id))
        except DoesNotExist:
            # Register new PID for the saved Blob.
            sub_url = reverse(
                "core_linked_records_provider_record",
                kwargs={
                    "provider": settings.ID_PROVIDER_SYSTEM_NAME,
                    "record": settings.ID_PROVIDER_PREFIX_BLOB,
                },
            )
            pid_response = send_post_request(
                f"{settings.SERVER_URI}{sub_url}?format=json"
            )
            blob_pid = json.loads(pid_response.content)["url"]

            blob_api.set_pid_for_blob(document.id, blob_pid)
    except Exception as exc:
        error_message = (
            f"An error occurred while setting a PID for blob '{document.id}'"
        )

        logger.error(f"{error_message}: {str(exc)}")
        raise CoreError(f"{error_message}.")



""" Permission tests for core_linked_records_app.rest.pid_xpath.views
"""
from unittest import TestCase

from rest_framework import status
from rest_framework.response import Response
from rest_framework_mongoengine.generics import (
    ListCreateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from unittest.mock import patch

from core_linked_records_app.rest.pid_xpath import views as pid_xpath_views
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock


class TestPidXpathListViewGet(TestCase):
    def test_anonymous_returns_403(self):
        response = RequestMock.do_request_get(
            pid_xpath_views.PidXpathListView.as_view(), None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_200(self):
        mock_user = create_mock_user("1")

        response = RequestMock.do_request_get(
            pid_xpath_views.PidXpathListView.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_staff_returns_200(self):
        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_get(
            pid_xpath_views.PidXpathListView.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestPidXpathListViewPost(TestCase):
    def test_anonymous_returns_403(self):
        response = RequestMock.do_request_post(
            pid_xpath_views.PidXpathListView.as_view(),
            None,
            data={"xpath": "mock.xpath", "template": "mock_template_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(ListCreateAPIView, "post")
    def test_authenticated_returns_201(self, mock_view_post):
        mock_user = create_mock_user("1")

        mock_view_post.return_value = Response(status=status.HTTP_201_CREATED)
        response = RequestMock.do_request_post(
            pid_xpath_views.PidXpathListView.as_view(),
            mock_user,
            data={"xpath": "mock.xpath", "template": "mock_template_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    @patch.object(ListCreateAPIView, "post")
    def test_staff_returns_201(self, mock_view_post):
        mock_user = create_mock_user("1", is_staff=True)

        mock_view_post.return_value = Response(status=status.HTTP_201_CREATED)
        response = RequestMock.do_request_post(
            pid_xpath_views.PidXpathListView.as_view(),
            mock_user,
            data={"xpath": "mock.xpath", "template": "mock_template_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)


class TestPidXpathDetailViewGet(TestCase):
    def test_anonymous_returns_403(self):
        response = RequestMock.do_request_get(
            pid_xpath_views.PidXpathDetailView.as_view(), None, param={"id": 0}
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(RetrieveUpdateDestroyAPIView, "get")
    def test_authenticated_returns_200(self, mock_view_get):
        mock_user = create_mock_user("1")

        mock_view_get.return_value = Response(status=status.HTTP_200_OK)
        response = RequestMock.do_request_get(
            pid_xpath_views.PidXpathDetailView.as_view(), mock_user, param={"id": 0}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(RetrieveUpdateDestroyAPIView, "get")
    def test_staff_returns_200(self, mock_view_get):
        mock_user = create_mock_user("1", is_staff=True)

        mock_view_get.return_value = Response(status=status.HTTP_200_OK)
        response = RequestMock.do_request_get(
            pid_xpath_views.PidXpathDetailView.as_view(), mock_user, param={"id": 0}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestPidXpathDetailViewPatch(TestCase):
    def test_anonymous_returns_403(self):
        response = RequestMock.do_request_patch(
            pid_xpath_views.PidXpathDetailView.as_view(),
            None,
            data={"xpath": "mock.xpath", "template": "mock_template_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(RetrieveUpdateDestroyAPIView, "patch")
    def test_authenticated_returns_200(self, mock_view_patch):
        mock_user = create_mock_user("1")

        mock_view_patch.return_value = Response(status=status.HTTP_200_OK)
        response = RequestMock.do_request_patch(
            pid_xpath_views.PidXpathDetailView.as_view(),
            mock_user,
            data={"xpath": "mock.xpath", "template": "mock_template_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(RetrieveUpdateDestroyAPIView, "patch")
    def test_staff_returns_200(self, mock_view_patch):
        mock_user = create_mock_user("1", is_staff=True)

        mock_view_patch.return_value = Response(status=status.HTTP_200_OK)
        response = RequestMock.do_request_patch(
            pid_xpath_views.PidXpathDetailView.as_view(),
            mock_user,
            data={"xpath": "mock.xpath", "template": "mock_template_id"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestPidXpathDetailViewDelete(TestCase):
    def test_anonymous_returns_403(self):
        response = RequestMock.do_request_delete(
            pid_xpath_views.PidXpathDetailView.as_view(),
            None,
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(RetrieveUpdateDestroyAPIView, "delete")
    def test_authenticated_returns_204(self, mock_view_delete):
        mock_user = create_mock_user("1")

        mock_view_delete.return_value = Response(status=status.HTTP_204_NO_CONTENT)
        response = RequestMock.do_request_delete(
            pid_xpath_views.PidXpathDetailView.as_view(),
            mock_user,
        )

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    @patch.object(RetrieveUpdateDestroyAPIView, "delete")
    def test_staff_returns_204(self, mock_view_delete):
        mock_user = create_mock_user("1", is_staff=True)

        mock_view_delete.return_value = Response(status=status.HTTP_204_NO_CONTENT)
        response = RequestMock.do_request_delete(
            pid_xpath_views.PidXpathDetailView.as_view(),
            mock_user,
        )

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)



""" Serializers for calls related to `PidXpath` model.
"""
from rest_framework_mongoengine import serializers

from core_linked_records_app.components.pid_xpath.models import PidXpath


class PidXpathSerializer(serializers.DocumentSerializer):
    class Meta:
        model = PidXpath



""" Unit tests for core_linked_records_app.components.pid_xpath.api
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.pid_xpath import api as pid_xpath_api
from core_linked_records_app.components.pid_xpath.models import PidXpath
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons.exceptions import ApiError
from core_main_app.components.template import (
    api as template_api,
)
from tests import mocks


class TestGetByTemplateId(TestCase):
    def setUp(self) -> None:
        self.mock_template_id = "mock_template_id"
        self.kwargs = {
            "template_id": self.mock_template_id,
            "request": mocks.MockRequest(),
        }

    @patch.object(template_api, "get_all")
    def test_get_all_template_failure_raises_api_error(self, mock_get_all):
        mock_get_all.side_effect = Exception("mock_get_all_exception")

        with self.assertRaises(ApiError):
            pid_xpath_api.get_by_template_id(**self.kwargs)

    @patch.object(template_api, "get_all")
    def test_template_id_not_accessible_raises_acl_error(self, mock_get_all):
        mock_get_all.return_value = []

        with self.assertRaises(AccessControlError):
            pid_xpath_api.get_by_template_id(**self.kwargs)

    @patch.object(PidXpath, "get_by_template_id")
    @patch.object(template_api, "get_all")
    def test_get_by_template_id_failure_raises_api_error(
        self, mock_get_all, mock_get_by_template_id
    ):
        mock_template = mocks.MockDocument()
        mock_template.pk = self.mock_template_id
        mock_get_all.return_value = [mock_template]
        mock_get_by_template_id.side_effect = Exception(
            "mock_get_by_template_id_exception"
        )

        with self.assertRaises(ApiError):
            pid_xpath_api.get_by_template_id(**self.kwargs)

    @patch.object(PidXpath, "__init__")
    @patch.object(PidXpath, "get_by_template_id")
    @patch.object(template_api, "get_all")
    def test_pid_xpath_init_failure_raises_api_error(
        self, mock_get_all, mock_get_by_template_id, mock_pid_xpath
    ):
        mock_template = mocks.MockDocument()
        mock_template.pk = self.mock_template_id
        mock_get_all.return_value = [mock_template]
        mock_get_by_template_id.return_value = None
        mock_pid_xpath.side_effect = Exception("mock_pid_xpath_exception")

        with self.assertRaises(ApiError):
            pid_xpath_api.get_by_template_id(**self.kwargs)

    @patch.object(PidXpath, "__init__")
    @patch.object(PidXpath, "get_by_template_id")
    @patch.object(template_api, "get_all")
    def test_returns_new_pid_xpath_if_not_exists(
        self, mock_get_all, mock_get_by_template_id, mock_pid_xpath
    ):
        expected_result = "mock_pid_xpath_object"
        mock_template = mocks.MockDocument()
        mock_template.pk = self.mock_template_id
        mock_get_all.return_value = [mock_template]
        mock_get_by_template_id.return_value = expected_result

        self.assertEquals(
            pid_xpath_api.get_by_template_id(**self.kwargs), expected_result
        )

    @patch.object(PidXpath, "__new__")
    @patch.object(PidXpath, "get_by_template_id")
    @patch.object(template_api, "get_all")
    def test_returns_pid_xpath_if_exists(
        self, mock_get_all, mock_get_by_template_id, mock_pid_xpath
    ):
        expected_result = "mock_pid_xpath"
        mock_template = mocks.MockDocument()
        mock_template.pk = self.mock_template_id
        mock_get_all.return_value = [mock_template]
        mock_get_by_template_id.return_value = None
        mock_pid_xpath.return_value = expected_result

        self.assertEquals(
            pid_xpath_api.get_by_template_id(**self.kwargs), expected_result
        )


class TestGetAll(TestCase):
    @patch.object(template_api, "get_all")
    def test_template_api_get_all_failure_raises_api_error(self, mock_get_all):
        mock_get_all.side_effect = Exception("mock_get_all_exception")

        with self.assertRaises(ApiError):
            pid_xpath_api.get_all(mocks.MockRequest())

    @patch.object(PidXpath, "get_all_by_template_list")
    @patch.object(template_api, "get_all")
    def test_get_all_by_template_list_failure_raises_api_error(
        self, mock_get_all, mock_get_all_by_template_list
    ):
        mock_get_all.return_value = []
        mock_get_all_by_template_list.side_effect = Exception("mock_get_all_exception")

        with self.assertRaises(ApiError):
            pid_xpath_api.get_all(mocks.MockRequest())

    @patch.object(PidXpath, "get_all_by_template_list")
    @patch.object(template_api, "get_all")
    def test_returns_get_all_by_template_list_output(
        self, mock_get_all, mock_get_all_by_template_list
    ):
        expected_result = "mock_get_all_by_template_list"
        mock_get_all.return_value = []
        mock_get_all_by_template_list.return_value = expected_result

        self.assertEquals(pid_xpath_api.get_all(mocks.MockRequest()), expected_result)



""" Unit tests for core_linked_records_app.rest.query.views
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.pid_xpath import api as pid_xpath_api
from core_linked_records_app.rest.query import views as query_views
from core_main_app.components.data import api as data_api
from core_main_app.rest.data.abstract_views import AbstractExecuteLocalQueryView
from tests import mocks


class TestExecuteLocalPIDQueryViewBuildQuery(TestCase):
    @patch.object(AbstractExecuteLocalQueryView, "build_query")
    @patch.object(pid_xpath_api, "get_all")
    def test_add_pid_query_if_init_query_has_and(
        self, mock_pid_xpath_get_all, mock_build_query
    ):
        mock_pid_xpath_get_all.return_value = []
        mock_build_query.return_value = {}

        test_view = query_views.ExecuteLocalPIDQueryView()
        test_view.request = mocks.MockRequest()
        result = test_view.build_query("mock_query")

        self.assertEquals(len(result["$and"]), 2)

    @patch.object(AbstractExecuteLocalQueryView, "build_query")
    @patch.object(pid_xpath_api, "get_all")
    def test_append_pid_query_if_init_query_has_no_and(
        self, mock_pid_xpath_get_all, mock_build_query
    ):
        mock_pid_xpath_get_all.return_value = []
        mock_build_query.return_value = {"$and": ["query1", "query2"]}

        test_view = query_views.ExecuteLocalPIDQueryView()
        test_view.request = mocks.MockRequest()
        result = test_view.build_query("mock_query")

        self.assertEquals(len(result["$and"]), 3)


class TestExecuteLocalPIDQueryViewExecuteRawQuery(TestCase):
    @patch.object(data_api, "execute_query")
    def test_no_data_returns_empty_list(self, mock_execute_query):
        mock_execute_query.return_value = []

        test_view = query_views.ExecuteLocalPIDQueryView()
        test_view.request = mocks.MockRequest()
        result = test_view.execute_raw_query("mock_query", "order_by_field")

        self.assertEquals(result, [])

    @patch.object(query_views, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(data_api, "execute_query")
    def test_returns_data_with_pid(
        self,
        mock_execute_query,
        mock_get_by_template_id,
        mock_get_dict_value_from_key_list,
    ):
        mock_data_pid = "mock_data_pid"
        mock_execute_query.return_value = [mocks.MockData() for _ in range(5)]
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.return_value = mock_data_pid
        expected_result = [{"pid": mock_data_pid} for _ in range(5)]

        test_view = query_views.ExecuteLocalPIDQueryView()
        test_view.request = mocks.MockRequest()
        result = test_view.execute_raw_query("mock_query", "order_by_field")

        self.assertEquals(result, expected_result)


class TestExecuteLocalPIDQueryViewBuildResponse(TestCase):
    def test_returns_list_of_data_pid(self):
        mock_data_pid = "mock_data_pid"
        mock_data_list = [{"pid": mock_data_pid} for _ in range(5)]
        expected_result = [mock_data_pid for _ in range(5)]

        test_view = query_views.ExecuteLocalPIDQueryView()
        response = test_view.build_response(mock_data_list)

        self.assertEquals(response.data, expected_result)



""" PID utilities related to data objects.
"""
from core_linked_records_app.utils.xml import (
    get_target_namespace_for_xsd_string,
    set_value_at_xpath,
    get_value_at_xpath,
    can_create_value_at_xpath,
)
from xml_utils.xpath import create_tree_from_xpath
from xml_utils.xsd_tree.xsd_tree import XSDTree


def set_pid_value_for_data(data, pid_xpath, pid_value):
    """Set the document PID into XML data and update `xml_content` in place.

    Args:
        data:
        pid_xpath:
        pid_value:
    """
    target_namespace = get_target_namespace_for_xsd_string(data.template.content)
    xml_tree = XSDTree.build_tree(data.xml_content)

    xml_tree = create_tree_from_xpath(pid_xpath, xml_tree, target_namespace)
    set_value_at_xpath(xml_tree, pid_xpath, pid_value, target_namespace)
    data.xml_content = XSDTree.tostring(xml_tree)

    # Update the whole document with the updated XML content
    data.convert_to_file()
    data.convert_to_dict()


def get_pid_value_for_data(data, pid_xpath):
    """Retrieve value located at `pid_xpath` of the data passed in parameter.

    Args:
        data:
        pid_xpath:

    Returns:
        str - Persitstent identifier
    """
    target_namespace = get_target_namespace_for_xsd_string(data.template.content)
    xml_tree = XSDTree.build_tree(data.xml_content)

    try:  # Get the PID from the `pid_xpath` value
        pid_value = get_value_at_xpath(xml_tree, pid_xpath, target_namespace)
        if type(pid_value) == str and pid_value.endswith(
            "/"
        ):  # Cleanup PID if it ends with a '/'
            pid_value = pid_value[:-1]
    except AssertionError:  # PID XPath not found in document
        assert can_create_value_at_xpath(
            data.xml_content, data.template.content, pid_xpath, "http://sample_pid.org"
        )
        pid_value = None

    return pid_value



""" Permission tests for core_linked_records_app.rest.query.views
"""
from unittest import TestCase

from rest_framework import status
from rest_framework.response import Response
from unittest.mock import patch

from core_linked_records_app.rest.query import views as query_views
from core_main_app.rest.data.abstract_views import AbstractExecuteLocalQueryView
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock


class TestExecuteLocalPIDQueryViewGet(TestCase):
    @patch.object(AbstractExecuteLocalQueryView, "execute_query")
    def test_anonymous_returns_200(self, mock_execute_query):
        mock_execute_query.return_value = Response(status=status.HTTP_200_OK)

        response = RequestMock.do_request_get(
            query_views.ExecuteLocalPIDQueryView.as_view(),
            None,
            data={"query": "{}"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(AbstractExecuteLocalQueryView, "execute_query")
    def test_authenticated_returns_200(self, mock_execute_query):
        mock_user = create_mock_user("1")
        mock_execute_query.return_value = Response(status=status.HTTP_200_OK)

        response = RequestMock.do_request_get(
            query_views.ExecuteLocalPIDQueryView.as_view(),
            mock_user,
            data={"query": "{}"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(AbstractExecuteLocalQueryView, "execute_query")
    def test_staff_returns_200(self, mock_execute_query):
        mock_user = create_mock_user("1", is_staff=True)
        mock_execute_query.return_value = Response(status=status.HTTP_200_OK)

        response = RequestMock.do_request_get(
            query_views.ExecuteLocalPIDQueryView.as_view(),
            mock_user,
            data={"query": "{}"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestExecuteLocalPIDQueryViewPost(TestCase):
    @patch.object(AbstractExecuteLocalQueryView, "execute_query")
    def test_anonymous_returns_200(self, mock_execute_query):
        mock_execute_query.return_value = Response(status=status.HTTP_200_OK)

        response = RequestMock.do_request_post(
            query_views.ExecuteLocalPIDQueryView.as_view(),
            None,
            data={"query": "{}"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(AbstractExecuteLocalQueryView, "execute_query")
    def test_authenticated_returns_200(self, mock_execute_query):
        mock_user = create_mock_user("1")
        mock_execute_query.return_value = Response(status=status.HTTP_200_OK)

        response = RequestMock.do_request_post(
            query_views.ExecuteLocalPIDQueryView.as_view(),
            mock_user,
            data={"query": "{}"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(AbstractExecuteLocalQueryView, "execute_query")
    def test_staff_returns_200(self, mock_execute_query):
        mock_user = create_mock_user("1", is_staff=True)
        mock_execute_query.return_value = Response(status=status.HTTP_200_OK)

        response = RequestMock.do_request_post(
            query_views.ExecuteLocalPIDQueryView.as_view(),
            mock_user,
            data={"query": "{}"},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


# FIXME cannot import oai_pmh_harvester in INSTALLED_APPS
# class TestExecuteOaiPmhPIDQueryViewGet(TestCase):
#     def test_anonymous_returns_200(self):
#         pass
#
#     @patch.object(ExecuteQueryView, "execute_query")
#     def test_authenticated_returns_200(self, mock_execute_query):
#         mock_user = create_mock_user("1")
#         mock_execute_query.return_value = Response(status=status.HTTP_200_OK)
#
#         response = RequestMock.do_request_get(
#             query_views.ExecuteOaiPmhPIDQueryView.as_view(),
#             mock_user,
#             data={"query": "{}"},
#         )
#
#         self.assertEqual(response.status_code, status.HTTP_200_OK)
#
#     def test_staff_returns_200(self):
#         pass
#
#
# class TestExecuteOaiPmhPIDQueryViewPost(TestCase):
#     def test_anonymous_returns_200(self):
#         pass
#
#     @patch.object(ExecuteQueryView, "execute_query")
#     def test_authenticated_returns_200(self, mock_execute_query):
#         mock_user = create_mock_user("1")
#         mock_execute_query.return_value = Response(status=status.HTTP_200_OK)
#
#         response = RequestMock.do_request_post(
#             query_views.ExecuteOaiPmhPIDQueryView.as_view(),
#             mock_user,
#             data={"query": "{}"},
#         )
#
#         self.assertEqual(response.status_code, status.HTTP_200_OK)
#
#     def test_staff_returns_200(self):
#         pass



""" Rest API views to retrieve PID settings
"""
from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from core_linked_records_app import settings
from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_linked_records_app.rest.pid_settings.serializers import (
    PidSettingsSerializer,
)


class PidSettingsView(APIView):
    """Retrieve PID settings"""

    permission_classes = (IsAuthenticated,)

    def get(self, request):
        """Retrieve the settings for the PID system

        Args:
            request:

        Returns:
        """
        try:
            pid_settings = pid_settings_api.get()
            pid_settings_data = PidSettingsSerializer(pid_settings).data

            response_data = {
                "format": settings.PID_FORMAT,
                "system": settings.ID_PROVIDER_SYSTEM_NAME,
                "prefixes": settings.ID_PROVIDER_PREFIXES,
            }

            pid_settings_data.update(response_data)

            return Response(
                pid_settings_data,
                status=status.HTTP_200_OK,
            )
        except Exception as exc:
            return Response(
                {
                    "message": f"An unexpected error occurred while displaying "
                    f"PidSettings: {str(exc)}"
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

    def patch(self, request):
        """Update settings for the PID system. Current, only works for automatically
        setting PIDs (`auto_set_pid`).

        Args:
            request:

        Returns:
        """
        if not request.user.is_superuser:
            return Response(
                {"message": "Only a superuser can use this feature."},
                status=status.HTTP_403_FORBIDDEN,
            )

        try:
            pid_settings_serializer = PidSettingsSerializer(data=request.data)

            if not pid_settings_serializer.is_valid():
                return Response(
                    {"message": "Invalid data provided"},
                    status=status.HTTP_400_BAD_REQUEST,
                )

            pid_settings_serializer.update(
                pid_settings_api.get(), pid_settings_serializer.validated_data
            )

            return self.get(request)
        except Exception as exc:
            return Response(
                {
                    "message": f"An unexpected error occurred while modifying "
                    f"PidSettings: {str(exc)}"
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )



""" Mock classes for objects used within core_linked_records_app
"""
from bson import ObjectId
from rest_framework import status
from unittest.mock import Mock


class MockModule(Mock):
    def get_by_id(self, *args, **kwargs):
        return "MockModule.get_by_id"


class MockLocalId(Mock):
    record_name = "mock_record_name"
    record_object_class = "mock.record.object.class"
    record_object_id = "mock_object_id"


class MockDocument(Mock):
    id = ObjectId()
    pk = ObjectId()


class MockData(MockDocument):
    dict_content = "mock_dict_content"


class MockPidSettings(Mock):
    auto_set_pid = True


class MockPidXpath(MockDocument):
    xpath = "mock.xpath"
    template = ObjectId()


class MockResponse(Mock):
    status_code = status.HTTP_200_OK
    data = dict()
    content = "{}"
    json_data = None

    def json(self):
        return self.json_data


class MockRequest(Mock):
    user = None


class MockProviderManager(Mock):
    provider_lookup_url = "mock_provider_url"
    create_exc = None
    create_result = None
    update_exc = None
    update_result = None
    get_exc = None
    get_result = None
    delete_exc = None
    delete_result = None

    def create(self, *args, **kwargs):
        if self.create_exc:
            raise self.create_exc

        return self.create_result

    def update(self, *args, **kwargs):
        if self.update_exc:
            raise self.update_exc

        return self.update_result

    def get(self, *args, **kwargs):
        if self.get_exc:
            raise self.get_exc

        return self.get_result

    def delete(self, *args, **kwargs):
        if self.delete_exc:
            raise self.delete_exc

        return self.delete_result


class MockQuery(Mock):
    data_sources = list()
    content = ""
    templates = list()


class MockDataSource(Mock):
    query_options = dict()
    order_by_field = ""
    capabilities = dict()


class MockAuthentication(Mock):
    type = ""


class MockSerializer(Mock):
    data = dict()
    is_valid_exc = None
    is_valid_result = True
    update_exc = None
    update_result = None

    def is_valid(self):
        if self.is_valid_exc:
            raise self.is_valid_exc
        return self.is_valid_result

    def update(self, *args):
        if self.update_exc:
            raise self.update_exc
        return self.update_result


class MockInstance(Mock):
    endpoint = "mock_endpoint"
    access_token = "mock_access_token"


class MockBlob(Mock):
    blob = "mock_blob"
    filename = "mock_filename"



""" Unit tests for core_linked_records_app.components.local_id.models
"""
from unittest import TestCase

from mongoengine import errors as mongoengine_errors
from unittest.mock import patch

from core_linked_records_app.components.local_id.models import LocalId
from core_main_app.commons import exceptions


class TestGetByName(TestCase):
    @patch.object(LocalId, "objects")
    def test_local_id_get_does_not_exist_raises_does_not_exist_error(
        self, mock_objects
    ):
        mock_objects.get.side_effect = mongoengine_errors.DoesNotExist(
            "mock_objects_get_does_not_exist"
        )

        with self.assertRaises(exceptions.DoesNotExist):
            LocalId.get_by_name("mock_record_name")

    @patch.object(LocalId, "objects")
    def test_local_id_get_failure_raises_model_error(self, mock_objects):
        mock_objects.get.side_effect = Exception("mock_objects_get_exception")

        with self.assertRaises(exceptions.ModelError):
            LocalId.get_by_name("mock_record_name")

    @patch.object(LocalId, "objects")
    def test_returns_local_id_get_output(self, mock_objects):
        expected_result = "mock_get_by_name"
        mock_objects.get.return_value = expected_result

        self.assertEquals(LocalId.get_by_name("mock_record_name"), expected_result)


class TestGetByClassAndId(TestCase):
    @patch.object(LocalId, "objects")
    def test_local_id_get_does_not_exist_raises_does_not_exist_error(
        self, mock_objects
    ):
        mock_objects.get.side_effect = mongoengine_errors.DoesNotExist(
            "mock_objects_get_does_not_exist"
        )

        with self.assertRaises(exceptions.DoesNotExist):
            LocalId.get_by_class_and_id("mock_record_class", "mock_record_id")

    @patch.object(LocalId, "objects")
    def test_local_id_get_failure_raises_model_error(self, mock_objects):
        mock_objects.get.side_effect = Exception("mock_objects_get_exception")

        with self.assertRaises(exceptions.ModelError):
            LocalId.get_by_class_and_id("mock_record_class", "mock_record_id")

    @patch.object(LocalId, "objects")
    def test_returns_local_id_get_output(self, mock_objects):
        expected_result = "mock_get_by_name"
        mock_objects.get.return_value = expected_result

        self.assertEquals(
            LocalId.get_by_class_and_id("mock_record_class", "mock_record_id"),
            expected_result,
        )


class TestUpdate(TestCase):
    def setUp(self):
        self.mock_local_id = LocalId()

    @patch.object(LocalId, "save")
    def test_local_id_save_not_unique_error_raises_not_unique_error(self, mock_save):
        mock_save.side_effect = mongoengine_errors.NotUniqueError(
            "mock_save_not_unique_error"
        )

        with self.assertRaises(exceptions.NotUniqueError):
            LocalId.upsert(self.mock_local_id)

    @patch.object(LocalId, "save")
    def test_local_id_save_failure_raises_model_error(self, mock_save):
        mock_save.side_effect = Exception("mock_save_exception")

        with self.assertRaises(exceptions.ModelError):
            LocalId.upsert(self.mock_local_id)

    @patch.object(LocalId, "save")
    def test_returns_local_id_get_output(self, mock_save):
        expected_result = "mock_get_by_name"
        mock_save.return_value = expected_result

        self.assertEquals(
            LocalId.upsert(self.mock_local_id),
            expected_result,
        )



""" Handle system abstract class
"""
from importlib import import_module

import logging
from abc import ABC, abstractmethod
from django.urls import reverse
from rest_framework import status

from core_linked_records_app import settings
from core_main_app.commons import exceptions
from core_main_app.utils.requests_utils.requests_utils import send_post_request

logger = logging.getLogger(__name__)


class AbstractIdProvider(ABC):
    def __init__(
        self,
        provider_name,
        provider_lookup_url,
        provider_registration_url,
        username,
        password,
    ):
        self.local_url = "%s%s" % (
            settings.SERVER_URI,
            reverse(
                "core_linked_records_provider_record",
                kwargs={"provider": provider_name, "record": ""},
            ),
        )
        self.local_url = self.local_url[:-1]  # Removing the final /
        self.provider_lookup_url = (
            provider_lookup_url if provider_lookup_url else self.local_url
        )
        self.provider_registration_url = provider_registration_url
        self.auth_token = self.encode_token(username, password)

    @abstractmethod
    def encode_token(self, username, password):
        raise NotImplementedError()

    @abstractmethod
    def get(self, record):
        raise NotImplementedError()

    @abstractmethod
    def create(self, prefix, record=None):
        raise NotImplementedError()

    @abstractmethod
    def update(self, record):
        raise NotImplementedError()

    @abstractmethod
    def delete(self, record):
        raise NotImplementedError()


class ProviderManager(object):
    """Manage provider instances from a given provider name"""

    def __init__(self):
        self._provider_instance = None

        self.provider_name = settings.ID_PROVIDER_SYSTEM_NAME
        self.provider_config = settings.ID_PROVIDER_SYSTEM_CONFIG

    def get(self, provider_name):
        if self._provider_instance is None:
            # Retrieve class name and module path
            id_provider_classpath = self.provider_config["class"].split(".")
            id_provider_classname = id_provider_classpath[-1]
            id_provider_modpath = ".".join(id_provider_classpath[:-1])

            # Import module and class
            id_provider_module = import_module(id_provider_modpath)
            id_provider_class = getattr(id_provider_module, id_provider_classname)

            # Initialize handel system instance
            self._provider_instance = id_provider_class(
                provider_name, *settings.ID_PROVIDER_SYSTEM_CONFIG["args"]
            )

        return self._provider_instance

    def find_provider_from_pid(self, pid):
        provider = self.get(settings.ID_PROVIDER_SYSTEM_NAME)

        return (
            settings.ID_PROVIDER_SYSTEM_NAME
            if pid.startswith(provider.provider_lookup_url)
            else None
        )


def retrieve_provider_name(pid_value):
    """Retrieve name of the provider given a PID.

    Args:
        pid_value:

    Returns:
         str - Provider name
    """

    if pid_value is None or pid_value == "":  # PID field left blank
        # Select the default provider if no PID has been chosen.
        provider_name = settings.ID_PROVIDER_SYSTEM_NAME
    else:  # PID specified in document.
        provider_name = ProviderManager().find_provider_from_pid(pid_value)

    # Detect provider specified by PID. Raise an error if no matching providers are
    # found.
    if provider_name is None:
        raise exceptions.ModelError("Invalid PID provided (provider not found)")

    return provider_name


def register_pid_for_data_id(provider_name, pid_value, data_id):
    """Registers a data with a set PID to a given provider.

    Args:
        provider_name:
        pid_value:
        data_id:

    Returns:
        str - Persistent identifier
    """
    from core_linked_records_app.system import api as system_api

    provider_manager = ProviderManager()
    provider = provider_manager.get(provider_name)
    registration_url = pid_value.replace(
        provider.provider_lookup_url, provider.local_url
    )

    document_pid_response = send_post_request("%s?format=json" % registration_url)

    # If an error happened during PID registration, try to relay the message from the
    # provider, otherwise relay a default error message.
    if document_pid_response.status_code == status.HTTP_500_INTERNAL_SERVER_ERROR:
        default_error_message = "An error occurred while creating the PID"
        try:
            raise exceptions.ModelError(
                document_pid_response.json().get("message", default_error_message)
            )
        except Exception as exc:  # If the response is not JSON parsable
            logger.error(f"{default_error_message}: {str(exc)}")
            raise exceptions.ModelError(default_error_message)

    if (
        document_pid_response.status_code != status.HTTP_201_CREATED
        and document_pid_response.status_code != status.HTTP_200_OK
        and system_api.get_data_by_pid(pid_value).pk != data_id
    ):
        raise exceptions.ModelError("Invalid PID provided")

    return document_pid_response.json()["url"]



""" Unit tests for core_linked_records_app.components.data.watch
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.data import watch as data_watch
from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_linked_records_app.system import api as system_api
from core_linked_records_app.utils import data as data_utils
from core_linked_records_app.utils import providers as providers_utils
from core_main_app.commons import exceptions
from tests import mocks


class TestSetDataPid(TestCase):
    def setUp(self):
        mock_sender = None
        mock_document = mocks.MockData()

        self.mock_kwargs = {"sender": mock_sender, "document": mock_document}

    @patch.object(pid_settings_api, "get")
    def test_pid_settings_get_failure_raises_core_error(self, mock_pid_settings_get):
        mock_pid_settings_get.side_effect = Exception("mock_pid_settings_get_exception")

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_get_pid_xpath_by_template_id_failure_raises_core_error(
        self, mock_pid_settings_get, mock_get_pid_xpath_by_template_id
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.side_effect = Exception(
            "mock_get_pid_xpath_by_template_id_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_get_xpath_from_dot_notation_failure_raises_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.side_effect = Exception(
            "mock_get_xpath_from_dot_notation_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_get_xpath_with_target_namespace_failure_raises_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.side_effect = Exception(
            "mock_get_xpath_with_target_namespace_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_get_pid_value_for_data_failure_returns_none(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.side_effect = Exception(
            "mock_get_pid_value_for_data_exception"
        )

        self.assertIsNone(data_watch.set_data_pid(**self.mock_kwargs))

    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_delete_pid_for_data_raises_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = "mock_pid_value"
        mock_delete_pid_for_data.side_effect = Exception(
            "mock_delete_pid_for_data_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_retrieve_provider_name_failure_raise_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = "mock_pid_value"
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.side_effect = Exception(
            "mock_retrieve_provider_name_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_provider_manager_get_failure_raise_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_is_pid_defined_failure_raise_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.side_effect = Exception(
            "mock_provider_manager_get_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(system_api, "is_pid_defined_for_document")
    @patch.object(system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_is_pid_defined_for_document_failure_raise_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_document,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_document.side_effect = Exception(
            "mock_is_pid_defined_for_document_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(system_api, "is_pid_defined_for_document")
    @patch.object(system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_pid_already_defined_raise_model_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_document,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_document.return_value = False

        with self.assertRaises(exceptions.ModelError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(providers_utils, "register_pid_for_data_id")
    @patch.object(system_api, "is_pid_defined_for_document")
    @patch.object(system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_register_pid_for_data_id_failure_raise_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_document,
        mock_register_pid_for_data_id,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_document.return_value = True
        mock_register_pid_for_data_id.side_effect = Exception(
            "mock_register_pid_for_data_id_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(data_utils, "set_pid_value_for_data")
    @patch.object(providers_utils, "register_pid_for_data_id")
    @patch.object(system_api, "is_pid_defined_for_document")
    @patch.object(system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_set_pid_value_for_data_failure_raise_core_error(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_document,
        mock_register_pid_for_data_id,
        mock_set_pid_value_for_data,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_document.return_value = True
        mock_register_pid_for_data_id.return_value = "mock_pid_value"
        mock_set_pid_value_for_data.side_effect = Exception(
            "mock_set_pid_value_for_data_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            data_watch.set_data_pid(**self.mock_kwargs)

    @patch.object(data_utils, "set_pid_value_for_data")
    @patch.object(providers_utils, "register_pid_for_data_id")
    @patch.object(system_api, "is_pid_defined_for_document")
    @patch.object(system_api, "is_pid_defined")
    @patch.object(providers_utils.ProviderManager, "get")
    @patch.object(providers_utils, "retrieve_provider_name")
    @patch.object(system_api, "delete_pid_for_data")
    @patch.object(data_utils, "get_pid_value_for_data")
    @patch.object(data_watch, "get_xpath_with_target_namespace")
    @patch.object(data_watch, "get_xpath_from_dot_notation")
    @patch.object(system_api, "get_pid_xpath_by_template_id")
    @patch.object(pid_settings_api, "get")
    def test_default_execution_returns_none(
        self,
        mock_pid_settings_get,
        mock_get_pid_xpath_by_template_id,
        mock_get_xpath_from_dot_notation,
        mock_get_xpath_with_target_namespace,
        mock_get_pid_value_for_data,
        mock_delete_pid_for_data,
        mock_retrieve_provider_name,
        mock_provider_manager_get,
        mock_is_pid_defined,
        mock_is_pid_defined_for_document,
        mock_register_pid_for_data_id,
        mock_set_pid_value_for_data,
    ):
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_get_pid_xpath_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_xpath_from_dot_notation.return_value = "mock_pid_xpath"
        mock_get_xpath_with_target_namespace.return_value = "mock_pid_xpath"
        mock_get_pid_value_for_data.return_value = None
        mock_delete_pid_for_data.return_value = None
        mock_retrieve_provider_name.return_value = "mock_provider_name"
        mock_provider_manager_get.return_value = mocks.MockProviderManager()
        mock_is_pid_defined.return_value = True
        mock_is_pid_defined_for_document.return_value = True
        mock_register_pid_for_data_id.return_value = "mock_pid_value"
        mock_set_pid_value_for_data.return_value = None

        self.assertIsNone(data_watch.set_data_pid(**self.mock_kwargs))



""" Linked records PID XPath objects.
"""
import logging
from django_mongoengine import Document
from mongoengine import errors as mongoengine_errors
from mongoengine import fields

from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import ModelError
from core_main_app.components.template.models import Template

logger = logging.getLogger(__name__)


class PidXpath(Document):
    xpath = fields.StringField(required=True)
    template = fields.ReferenceField(Template, required=True, unique=True)

    @staticmethod
    def get_all():
        try:
            return PidXpath.objects.all()
        except Exception as exc:
            raise ModelError(str(exc))

    @staticmethod
    def get_all_by_template_list(template_list):
        try:
            return PidXpath.objects.filter(template__in=template_list)
        except Exception as exc:
            raise ModelError(str(exc))

    @staticmethod
    def get_by_template_id(template_id):
        try:
            return PidXpath.objects.get(template=template_id)
        except mongoengine_errors.DoesNotExist:
            return None
        except Exception as exc:
            raise exceptions.ModelError(str(exc))



""" Linked records REST views
"""
import json
import logging
import re
from rest_framework import status
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.views import APIView

from core_linked_records_app import settings
from core_linked_records_app.components.blob.api import get_blob_by_pid
from core_linked_records_app.components.data.api import get_data_by_pid
from core_linked_records_app.rest.data.renderers.data_html_user_renderer import (
    DataHtmlUserRenderer,
)
from core_linked_records_app.rest.data.renderers.data_xml_renderer import (
    DataXmlRenderer,
)
from core_linked_records_app.utils.providers import ProviderManager
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.rest.data.serializers import DataSerializer
from core_main_app.utils.file import get_file_http_response

logger = logging.getLogger(__name__)


class ProviderRecordView(APIView):
    parser_classes = (JSONParser,)
    renderer_classes = (DataHtmlUserRenderer, JSONRenderer, DataXmlRenderer)

    def __init__(self):
        self.provider_manager = ProviderManager()
        super().__init__()

    def post(self, request, provider, record):
        """Create a handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        try:
            # Parse record entry to split between prefix and record ID.
            prefix_record_list = record.split("/")

            if len(prefix_record_list) == 1:  # No record provided
                prefix = prefix_record_list[0]
                record = None
            elif prefix_record_list[-1] == "":  # Only prefix provided
                prefix = "/".join(prefix_record_list[:-1])
                record = None
            else:  # Prefix and record provided
                prefix = "/".join(prefix_record_list[:-1])
                record = prefix_record_list[-1]

            # Assign default prefix if the prefix is undefined or not in the list of
            # authorized ones.
            if prefix == "" or prefix not in settings.ID_PROVIDER_PREFIXES:
                return Response(
                    {
                        "record": "/".join(prefix_record_list),
                        "url": request.build_absolute_uri("?"),
                        "message": "Invalid prefix specified",
                    },
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                )

            if (
                record is not None
                and record != ""
                and re.match(r"^(%s|)$" % settings.PID_FORMAT, record) is None
            ):
                return Response(
                    {
                        "record": "/".join(prefix_record_list),
                        "url": request.build_absolute_uri("?"),
                        "message": "Invalid record format",
                    },
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                )

            id_provider = self.provider_manager.get(provider)
            provider_response = id_provider.create(prefix, record)

            provider_content = json.loads(provider_response.content)
            return Response(provider_content, status=provider_response.status_code)
        except Exception as exc:
            return Response(
                {
                    "record": record,
                    "url": request.build_absolute_uri("?"),
                    "message": str(exc),
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

    def put(self, request, provider, record):
        """Update the value of a given handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        try:
            id_provider = self.provider_manager.get(provider)
            provider_response = id_provider.update(record)

            provider_content = json.loads(provider_response.content)

            return Response(provider_content, status=provider_response.status_code)
        except Exception as exc:
            return Response(
                {
                    "message": f"An unexpected error occurred while updating "
                    f"record: {str(exc)}"
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

    def get(self, request, provider, record):
        """Retrieve the local data of a given handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        try:
            id_provider = self.provider_manager.get(provider)
            provider_response = id_provider.get(record)

            try:
                query_result = get_data_by_pid(
                    json.loads(provider_response.content)["url"], request
                )
                return Response(
                    DataSerializer(query_result).data, status=status.HTTP_200_OK
                )
            except DoesNotExist:
                # Try to find PID in blobs
                try:
                    query_result = get_blob_by_pid(
                        json.loads(provider_response.content)["url"], request.user
                    )

                    return get_file_http_response(
                        query_result.blob, query_result.filename
                    )
                except AccessControlError as e:
                    content = {"message": str(e)}
                    return Response(content, status=status.HTTP_403_FORBIDDEN)
                except DoesNotExist:
                    content = {
                        "status": "error",
                        "code": status.HTTP_404_NOT_FOUND,
                        "message": "No document with specified handle found",
                    }
                    return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as exc:
            content = {
                "status": "error",
                "code": status.HTTP_500_INTERNAL_SERVER_ERROR,
                "message": str(exc),
            }
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def delete(self, request, provider, record):
        """Delete a handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        try:
            id_provider = self.provider_manager.get(provider)
            provider_response = id_provider.delete(record)

            provider_content = json.loads(provider_response.content)

            return Response(provider_content, status=provider_response.status_code)
        except Exception as exc:
            return Response(
                {
                    "message": f"An unexpected error occurred while deleting "
                    f"record: {str(exc)}"
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )



""" Unit tests for core_linked_records_app.components.blob.watch
"""
from unittest import TestCase

import json
from unittest.mock import patch

from core_linked_records_app.components.blob import api as blob_api
from core_linked_records_app.components.blob import watch as blob_watch
from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_main_app.commons import exceptions
from tests import mocks


class TestSetBlobPid(TestCase):
    def setUp(self):
        self.mock_document = mocks.MockDocument()

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(pid_settings_api, "get")
    def test_pid_setting_get_failure_raises_core_error(
        self, mock_get, mock_get_pid_for_blob
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_get.side_effect = Exception("mock_get_exception")

        with self.assertRaises(exceptions.CoreError):
            blob_watch.set_blob_pid(None, self.mock_document)

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_watch, "reverse")
    @patch.object(pid_settings_api, "get")
    def test_reverse_failure_raises_core_error(
        self, mock_get, mock_reverse, mock_get_pid_for_blob
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_get.return_value = mocks.MockPidSettings()
        mock_reverse.side_effect = Exception("mock_reverse_exception")

        with self.assertRaises(exceptions.CoreError):
            blob_watch.set_blob_pid(None, self.mock_document)

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_watch, "send_post_request")
    @patch.object(blob_watch, "reverse")
    @patch.object(pid_settings_api, "get")
    def test_send_post_request_failure_raises_core_error(
        self, mock_get, mock_reverse, mock_send_post_request, mock_get_pid_for_blob
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_get.return_value = mocks.MockPidSettings()
        mock_reverse.return_value = "mock/reverse/url"
        mock_send_post_request.side_effect = Exception(
            "mock_send_post_request_exception"
        )

        with self.assertRaises(exceptions.CoreError):
            blob_watch.set_blob_pid(None, self.mock_document)

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_watch, "send_post_request")
    @patch.object(blob_watch, "reverse")
    @patch.object(pid_settings_api, "get")
    def test_json_loads_failure_raises_core_error(
        self, mock_get, mock_reverse, mock_send_post_request, mock_get_pid_for_blob
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_get.return_value = mocks.MockPidSettings()
        mock_reverse.return_value = "mock/reverse/url"
        mock_send_post_request.return_value = "mock_not_json_return_value"

        with self.assertRaises(exceptions.CoreError):
            blob_watch.set_blob_pid(None, self.mock_document)

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_api, "set_pid_for_blob")
    @patch.object(blob_watch, "send_post_request")
    @patch.object(blob_watch, "reverse")
    @patch.object(pid_settings_api, "get")
    def test_set_pid_for_blob_failure_raises_core_error(
        self,
        mock_get,
        mock_reverse,
        mock_send_post_request,
        mock_set_pid_for_blob,
        mock_get_pid_for_blob,
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_get.return_value = mocks.MockPidSettings()
        mock_reverse.return_value = "mock/reverse/url"

        mock_pid_response = mocks.MockResponse()
        mock_pid_response.content = json.dumps({"url": "mock_pid_response_url"})
        mock_send_post_request.return_value = mock_pid_response
        mock_set_pid_for_blob.side_effect = Exception("mock_set_pid_for_blob_exception")

        with self.assertRaises(exceptions.CoreError):
            blob_watch.set_blob_pid(None, self.mock_document)

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_api, "set_pid_for_blob")
    @patch.object(pid_settings_api, "get")
    def test_auto_set_pid_false_does_not_assign_pid(
        self, mock_get, mock_set_pid_for_blob, mock_get_pid_for_blob
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_pid_settings = mocks.MockPidSettings()
        mock_pid_settings.auto_set_pid = False
        mock_get.return_value = mock_pid_settings

        blob_watch.set_blob_pid(None, self.mock_document)
        mock_set_pid_for_blob.assert_not_called()

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_api, "set_pid_for_blob")
    @patch.object(blob_watch, "send_post_request")
    @patch.object(blob_watch, "reverse")
    @patch.object(pid_settings_api, "get")
    def test_auto_set_pid_true_assign_pid(
        self,
        mock_get,
        mock_reverse,
        mock_send_post_request,
        mock_set_pid_for_blob,
        mock_get_pid_for_blob,
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_get.return_value = mocks.MockPidSettings()
        mock_reverse.return_value = "mock/reverse/url"

        mock_pid_response = mocks.MockResponse()
        mock_pid_response.content = json.dumps({"url": "mock_pid_response_url"})
        mock_send_post_request.return_value = mock_pid_response
        mock_set_pid_for_blob.return_value = None

        blob_watch.set_blob_pid(None, self.mock_document)
        mock_set_pid_for_blob.assert_called_once()

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(pid_settings_api, "get")
    def test_auto_set_pid_false_returns_none(self, mock_get, mock_get_pid_for_blob):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_pid_settings = mocks.MockPidSettings()
        mock_pid_settings.auto_set_pid = False
        mock_get.return_value = mock_pid_settings

        result = blob_watch.set_blob_pid(None, self.mock_document)
        self.assertIsNone(result)

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_api, "set_pid_for_blob")
    @patch.object(blob_watch, "send_post_request")
    @patch.object(blob_watch, "reverse")
    @patch.object(pid_settings_api, "get")
    def test_auto_set_pid_true_returns_none(
        self,
        mock_get,
        mock_reverse,
        mock_send_post_request,
        mock_set_pid_for_blob,
        mock_get_pid_for_blob,
    ):
        mock_get_pid_for_blob.side_effect = exceptions.DoesNotExist(
            "pid_does_not_exist"
        )
        mock_get.return_value = mocks.MockPidSettings()
        mock_reverse.return_value = "mock/reverse/url"

        mock_pid_response = mocks.MockResponse()
        mock_pid_response.content = json.dumps({"url": "mock_pid_response_url"})
        mock_send_post_request.return_value = mock_pid_response
        mock_set_pid_for_blob.return_value = None

        result = blob_watch.set_blob_pid(None, self.mock_document)
        self.assertIsNone(result)

    @patch.object(blob_api, "get_pid_for_blob")
    @patch.object(blob_api, "set_pid_for_blob")
    def test_pid_not_created_if_already_assigned(
        self, mock_set_pid_for_blob, mock_get_pid_for_blob
    ):
        mock_get_pid_for_blob.return_value = "mock_pid"

        blob_watch.set_blob_pid(None, self.mock_document)
        self.assertFalse(mock_set_pid_for_blob.called)



""" Signals to trigger before Data save
"""
from os.path import join

import logging

from core_linked_records_app import settings
from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_linked_records_app.system import api as system_api
from core_linked_records_app.utils import data as data_utils
from core_linked_records_app.utils.xml import (
    get_xpath_from_dot_notation,
    get_xpath_with_target_namespace,
)
from core_main_app.commons import exceptions
from core_main_app.components.data.models import Data
from signals_utils.signals.mongo import signals, connector

logger = logging.getLogger(__name__)


def init():
    """Connect to Data object events."""
    connector.connect(set_data_pid, signals.pre_save, sender=Data)


def set_data_pid(sender, document, **kwargs):
    """Set the PID in the XML field specified in the settings. If the PID
    already exists and is valid, it is not reset.

    Params:
        sender:
        document:
        kwargs:

    Returns:
    """
    from core_linked_records_app.utils import providers as providers_utils

    try:
        if not pid_settings_api.get().auto_set_pid:
            return

        # Retrieve PID XPath from `PidSettings.xpath_list`. Skip PID assignment
        # if the PID XPath is not defined for the template.
        template_pid_xpath = system_api.get_pid_xpath_by_template_id(
            document.template.pk
        )
        pid_xpath = get_xpath_with_target_namespace(
            get_xpath_from_dot_notation(template_pid_xpath.xpath),
            document.template.content,
        )

        try:  # Retrieve the PID located at predefined XPath
            pid_value = data_utils.get_pid_value_for_data(document, pid_xpath)
        except Exception as exc:  # XPath is not valid for current document
            logger.warning(
                f"Cannot create PID at {pid_xpath} for data {document.pk}: {str(exc)}"
            )
            return

        # Remove previous document PID from DB.
        if document.pk is not None:
            system_api.delete_pid_for_data(document)

        provider_name = providers_utils.retrieve_provider_name(pid_value)

        # Assign default value for undefined PID and check that the PID is not
        # defined for a document other than the current document.
        if pid_value is None or pid_value == "":
            pid_value = join(
                providers_utils.ProviderManager()
                .get(provider_name)
                .provider_lookup_url,
                settings.ID_PROVIDER_PREFIX_DEFAULT,
            )

        if system_api.is_pid_defined(pid_value) and (
            document.pk is None
            or not system_api.is_pid_defined_for_document(pid_value, document.pk)
        ):
            raise exceptions.ModelError("PID already defined for another document")

        # Register PID and write resulting URL in document
        pid_value = providers_utils.register_pid_for_data_id(
            provider_name, pid_value, document.pk
        )
        data_utils.set_pid_value_for_data(document, pid_xpath, pid_value)
    except exceptions.ModelError as model_error:
        logger.error(f"An error occurred while assigning PID: {str(model_error)}")
        raise exceptions.ModelError(str(model_error))
    except Exception as exc:
        if not document.id:
            data_definition = "new data"
        else:
            data_definition = f"data '{document.id}'"

        error_message = f"An error occurred while assigning PID to {data_definition}"

        logger.error(f"{error_message}: {str(exc)}")
        raise exceptions.CoreError(f"{error_message}.")



""" Unit tests for core_linked_records_app.utils.blob
"""
from unittest import TestCase


class TestGetBlobDownloadRegex(TestCase):
    def test_no_pid_in_document_returns_empty_list(self):
        pass

    def test_returned_list_has_correct_length(self):
        pass



""" Local resolver API
"""
import logging

from core_linked_records_app.components.pid_xpath import api as pid_xpath_api
from core_linked_records_app.utils.dict import get_dict_value_from_key_list
from core_main_app.commons.exceptions import ApiError
from core_oaipmh_harvester_app.components.oai_record import api as oai_record_data

logger = logging.getLogger(__name__)


def get_pid_for_data(oai_record_id, request):
    """Retrieve PID matching the document ID provided.

    Args:
        oai_record_id:
        request: HttpRequest

    Returns:
    """
    try:
        # Retrieve the document passed as input and extra the PID field.
        data = oai_record_data.get_by_id(oai_record_id, request.user)

        pid_xpath_object = pid_xpath_api.get_by_template_id(
            data.harvester_metadata_format.template.pk, request
        )
        pid_xpath = pid_xpath_object.xpath

        # Return PID value from the document and the PID_XPATH
        return get_dict_value_from_key_list(data.dict_content, pid_xpath.split("."))
    except Exception as exc:
        error_message = "An unexpected error occurred while retrieving PID for OAI data"

        logger.error(f"{error_message}: {str(exc)}")
        raise ApiError(f"{error_message}.")



""" API for PidXpath model
"""
import logging

from core_linked_records_app import settings
from core_linked_records_app.components.pid_xpath.models import PidXpath
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons.exceptions import ApiError
from core_main_app.components.template import (
    api as template_api,
)

logger = logging.getLogger(__name__)


def get_by_template_id(template_id, request):
    """Retrieve XPath associated with a specific template ID

    Args:
        template_id: ObjectId
        request: HttpRequest

    Returns:
        str - XPath for the given template ID
    """
    try:
        if template_id not in [
            template.pk for template in template_api.get_all(request=request)
        ]:
            raise AccessControlError("Template not accessible to the current user")

        pid_xpath_object = PidXpath.get_by_template_id(template_id)

        if pid_xpath_object is None:
            return PidXpath(template=template_id, xpath=settings.PID_XPATH)
        else:
            return pid_xpath_object
    except AccessControlError as ace:
        raise AccessControlError(str(ace))
    except Exception as exc:
        error_message = (
            f"An unexpected error occurred while retrieving PidXpath "
            f"assigned to template {template_id}"
        )

        logger.error(f"{error_message}: {str(exc)}")
        raise ApiError(str(exc))


def get_all(request):
    """Retrieve all XPath stored in `PidSettings`

    Args:
        request: HttpRequest

    Returns:
        list<str> - List of XPath
    """
    try:
        return PidXpath.get_all_by_template_list(
            [template.pk for template in template_api.get_all(request=request)]
        )
    except Exception as exc:
        error_message = f"An unexpected error occurred while retrieving all PidXpath"

        logger.error(f"{error_message}: {str(exc)}")
        raise ApiError(str(exc))



""" Initialization function for PID Settings
"""
import logging

from core_linked_records_app import settings
from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_linked_records_app.components.pid_settings.models import PidSettings
from core_main_app.commons.exceptions import CoreError

logger = logging.getLogger(__name__)


def init():
    try:
        if not PidSettings.get():
            pid_settings = PidSettings(auto_set_pid=settings.AUTO_SET_PID)
            pid_settings_api.upsert(pid_settings)
    except Exception as exc:
        error_message = "An unexpected error occurred while initializing PidSettings"

        logger.error(f"{error_message}: {str(exc)}")
        raise CoreError(f"{error_message}.")



""" Unit tests for core_linked_records_app.components.pid_settings.watch
"""
from unittest import TestCase

from unittest.mock import patch

from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_linked_records_app.components.pid_settings import watch as pid_settings_watch
from core_linked_records_app.components.pid_settings.models import PidSettings
from core_main_app.commons.exceptions import CoreError


class TestInit(TestCase):
    @patch.object(PidSettings, "get")
    def test_pid_settings_get_failure_raises_core_error(self, mock_pid_settings_get):
        mock_pid_settings_get.side_effect = Exception("mock_pid_settings_get_exception")

        with self.assertRaises(CoreError):
            pid_settings_watch.init()

    @patch.object(PidSettings, "get")
    def test_pid_settings_get_true_returns_none(self, mock_pid_settings_get):
        mock_pid_settings_get.return_value = True

        self.assertIsNone(pid_settings_watch.init())

    @patch.object(PidSettings, "__init__")
    @patch.object(PidSettings, "get")
    def test_pid_settings_init_raises_core_error(
        self, mock_pid_settings_get, mock_pid_settings_init
    ):
        mock_pid_settings_get.return_value = False
        mock_pid_settings_init.side_effect = Exception(
            "mock_pid_settings_init_exception"
        )

        with self.assertRaises(CoreError):
            pid_settings_watch.init()

    @patch.object(pid_settings_api, "upsert")
    @patch.object(PidSettings, "__init__")
    @patch.object(PidSettings, "get")
    def test_pid_settings_upsert_failure_raises_core_error(
        self, mock_pid_settings_get, mock_pid_settings_init, mock_pid_settings_upsert
    ):
        mock_pid_settings_get.return_value = False
        mock_pid_settings_init.return_value = None
        mock_pid_settings_upsert.side_effect = Exception(
            "mock_pid_settings_upsert_exception"
        )

        with self.assertRaises(CoreError):
            pid_settings_watch.init()

    @patch.object(pid_settings_api, "upsert")
    @patch.object(PidSettings, "__init__")
    @patch.object(PidSettings, "get")
    def test_pid_settings_get_false_returns_none(
        self, mock_pid_settings_get, mock_pid_settings_init, mock_pid_settings_upsert
    ):
        mock_pid_settings_get.return_value = False
        mock_pid_settings_init.return_value = None
        mock_pid_settings_upsert.return_value = None

        self.assertIsNone(pid_settings_watch.init())



""" Unit tests for core_linked_records_app.components.data.api
"""
from unittest import TestCase

from django.http import HttpRequest
from unittest.mock import patch, Mock

from core_linked_records_app.components.data import api as pid_data_api
from core_linked_records_app.components.pid_xpath import api as pid_xpath_api
from core_main_app.commons import exceptions
from core_main_app.components.data import api as main_data_api
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests import mocks


class TestGetDataByPid(TestCase):
    def setUp(self) -> None:
        mock_pid = "mock_pid"
        mock_user = create_mock_user("1")
        mock_request = Mock(spec=HttpRequest)
        mock_request.user = mock_user

        self.mock_kwargs = {"pid": mock_pid, "request": mock_request}

    @patch.object(main_data_api, "execute_query")
    def test_execute_query_failure_raises_api_error(self, mock_execute_query):
        mock_execute_query.side_effect = Exception("mock_execute_query_exception")

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_data_by_pid(**self.mock_kwargs)

    @patch.object(main_data_api, "execute_query")
    def test_no_result_raise_does_not_exist_error(self, mock_execute_query):
        mock_execute_query.return_value = []

        with self.assertRaises(exceptions.DoesNotExist):
            pid_data_api.get_data_by_pid(**self.mock_kwargs)

    @patch.object(main_data_api, "execute_query")
    def test_several_results_raise_api_error(self, mock_execute_query):
        mock_execute_query.return_value = ["item_a", "item_b"]

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_data_by_pid(**self.mock_kwargs)

    @patch.object(main_data_api, "execute_query")
    def test_single_result_is_returned(self, mock_execute_query):
        expected_result = "single_item"
        mock_execute_query.return_value = [expected_result]

        result = pid_data_api.get_data_by_pid(**self.mock_kwargs)
        self.assertEquals(result, expected_result)


class TestGetPidsForDataList(TestCase):
    def setUp(self) -> None:
        mock_data_id_list = ["mock_id_a", "mock_id_b"]
        mock_user = create_mock_user("1")
        mock_request = Mock(spec=HttpRequest)
        mock_request.user = mock_user

        self.mock_kwargs = {"data_id_list": mock_data_id_list, "request": mock_request}

    @patch.object(main_data_api, "get_by_id_list")
    def test_get_by_id_list_failure_raises_api_error(self, mock_get_by_id_list):
        mock_get_by_id_list.side_effect = Exception("mock_get_by_id_list_exception")

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_pids_for_data_list(**self.mock_kwargs)

    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(main_data_api, "get_by_id_list")
    def test_get_by_template_id_failure_raises_api_error(
        self,
        mock_get_by_id_list,
        mock_get_by_template_id,
    ):
        mock_get_by_id_list.return_value = [mocks.MockData()]
        mock_get_by_template_id.side_effect = Exception(
            "mock_get_by_template_id_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_pids_for_data_list(**self.mock_kwargs)

    @patch.object(pid_data_api, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(main_data_api, "get_by_id_list")
    def test_get_dict_value_from_key_list_failure_raises_api_error(
        self,
        mock_get_by_id_list,
        mock_get_by_template_id,
        mock_get_dict_value_from_key_list,
    ):
        mock_get_by_id_list.return_value = [mocks.MockData()]
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.side_effect = Exception(
            "mock_get_dict_value_from_key_list_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_pids_for_data_list(**self.mock_kwargs)

    @patch.object(pid_data_api, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(main_data_api, "get_by_id_list")
    def test_list_of_pid_does_not_contain_none_values(
        self,
        mock_get_by_id_list,
        mock_get_by_template_id,
        mock_get_dict_value_from_key_list,
    ):
        mock_get_by_id_list.return_value = [mocks.MockData()]
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.return_value = None

        result = pid_data_api.get_pids_for_data_list(**self.mock_kwargs)
        self.assertEquals(result, [])

    @patch.object(pid_data_api, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(main_data_api, "get_by_id_list")
    def test_returns_list_of_pid(
        self,
        mock_get_by_id_list,
        mock_get_by_template_id,
        mock_get_dict_value_from_key_list,
    ):
        mock_pid = "mock_pid"
        expected_result = [mock_pid for _ in range(5)]
        mock_get_by_id_list.return_value = [mocks.MockData() for _ in range(5)]
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.return_value = mock_pid

        result = pid_data_api.get_pids_for_data_list(**self.mock_kwargs)
        self.assertEquals(result, expected_result)


class TestGetPidForData(TestCase):
    def setUp(self) -> None:
        mock_data_id = mocks.MockData().pk
        mock_user = create_mock_user("1")
        mock_request = Mock(spec=HttpRequest)
        mock_request.user = mock_user

        self.mock_kwargs = {"data_id": mock_data_id, "request": mock_request}

        self.mock_global_data = mocks.MockData()

    @patch.object(main_data_api, "get_by_id")
    def test_get_by_id_failure_raises_api_error(self, mock_get_by_id):
        mock_get_by_id.side_effect = Exception("mock_get_by_id_exception")

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_pid_for_data(**self.mock_kwargs)

    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(main_data_api, "get_by_id")
    def test_get_by_template_id_failure_raises_api_error(
        self, mock_get_by_id, mock_get_by_template_id
    ):
        mock_get_by_id.return_value = self.mock_global_data
        mock_get_by_template_id.side_effect = Exception(
            "mock_get_by_template_id_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_pid_for_data(**self.mock_kwargs)

    @patch.object(pid_data_api, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(main_data_api, "get_by_id")
    def test_get_dict_value_from_key_list_failure_raises_api_error(
        self, mock_get_by_id, mock_get_by_template_id, mock_get_dict_value_from_key_list
    ):
        mock_get_by_id.return_value = self.mock_global_data
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.side_effect = Exception(
            "mock_get_dict_value_from_key_list_exception"
        )

        with self.assertRaises(exceptions.ApiError):
            pid_data_api.get_pid_for_data(**self.mock_kwargs)

    @patch.object(pid_data_api, "get_dict_value_from_key_list")
    @patch.object(pid_xpath_api, "get_by_template_id")
    @patch.object(main_data_api, "get_by_id")
    def test_returns_get_dict_value_from_key_list_output(
        self, mock_get_by_id, mock_get_by_template_id, mock_get_dict_value_from_key_list
    ):
        expected_result = "mock_pid"
        mock_get_by_id.return_value = self.mock_global_data
        mock_get_by_template_id.return_value = mocks.MockPidXpath()
        mock_get_dict_value_from_key_list.return_value = expected_result

        result = pid_data_api.get_pid_for_data(**self.mock_kwargs)
        self.assertEquals(result, expected_result)



""" Serializers for calls related to `PidSettings` model.
"""
from rest_framework_mongoengine import serializers

from core_linked_records_app.components.pid_settings import api as pid_settings_api
from core_linked_records_app.components.pid_settings.models import (
    PidSettings,
)


class PidSettingsSerializer(serializers.DocumentSerializer):
    class Meta:
        model = PidSettings

    def update(self, instance: PidSettings, validated_data):
        """Update a PidSetting instance

        Args:
            instance:
            validated_data:

        Returns:
            Updated PidSettings instance
        """
        instance.auto_set_pid = validated_data["auto_set_pid"]
        return pid_settings_api.upsert(instance)
